"""Constants used by various autofile modules"""

APP_NAME = "autofile"

# sentinel value to determine if a template did not match a field
NONE_STR_SENTINEL = f"__XYZZY_{APP_NAME}_TEMPLATE_NONE_XYZZY__"

# default value to use for none string in template
NONE_STR_DEFAULT = "_"

# default values for string manipulation template options
INPLACE_DEFAULT = ","



"""Plugin template for autofile to manipulate filepath """

import pathlib
import shlex
from typing import Iterable, List, Optional

import autofile
from autofile.path_utils import sanitize_dirname, sanitize_pathpart

FIELDS = {"{filepath}": "The full path to the file being processed"}

SUBFIELDS = {
    "name": "The name of the file",
    "stem": "The name of the file without the suffix (extension)",
    "suffix": "The suffix (extension) of the file, including the leading `.`",
    "parent": "The parent directory of the file",
}


@autofile.hookimpl
def get_template_help() -> Iterable:
    text = """
    The `{filepath}` fields returns the full path to the source file being processed. 
    Various attributes of the path can be accessed using "dot notation" (appended to the filepath field with a '.'). 
    For example, 
    `{filepath.name}` returns just the name of the file without the full path. 
    `{filepath.parent}` returns the parent directory of the file.
    
    Path attributes can be chained, for example `{filepath.parent.name}` returns just the name of the immediate parent directory without the full directory path.

    For example, if the field `{filepath}` is `'/Shared/files/IMG_1234.JPG'`:

    - `{filepath.parent}` is `'/Shared/files'`
    - `{filepath.name}` is `'IMG_1234.JPG'`
    - `{filepath.stem}` is `'IMG_1234'`
    - `{filepath.suffix}` is `'.JPG'`

    The following attributes are available:
    
    """
    fields = [["Field", "Description"], *[[k, v] for k, v in FIELDS.items()]]
    subfields = [["Subfield", "Description"], *[[k, v] for k, v in SUBFIELDS.items()]]
    return ["**File Path Fields**", fields, text, subfields]


@autofile.hookimpl
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """lookup value for template pathlib template fields

    Args:
        field: template field to find value for.

    Returns:
        The matching template value (which may be None).
    """
    field_stem = field.split(".")[0]
    if field_stem != "filepath":
        return None

    return [_get_pathlib_value(field, filepath)]


def _get_pathlib_value(field, value, quote=False):
    """Get the value for a pathlib.Path type template

    Args:
        field: the path field, e.g. "filename.stem"
        value: the value for the path component
        quote: bool; if true, quotes the returned path for safe execution in the shell
    """
    parts = field.split(".")

    if len(parts) == 1:
        return shlex.quote(value) if quote else value

    path = pathlib.Path(value)
    for attribute in parts[1:]:
        try:
            val = getattr(path, attribute)
            path = pathlib.Path(val)
        except AttributeError:
            raise ValueError(f"Illegal value for path template: {attribute}")

    val_str = str(val)
    if quote:
        val_str = shlex.quote(val_str)
    return val_str



""" plugin for autofile template: punctuation characters"""

from typing import Iterable, List, Optional

import autofile

PUNCTUATION = {
    "comma": ",",
    "semicolon": ";",
    "pipe": "|",
    "percent": "%",
    "openbrace": "{",
    "closebrace": "}",
    "openparens": "(",
    "closeparens": ")",
    "openbracket": "[",
    "closebracket": "]",
    "questionmark": "?",
    "newline": "\n",
    "lf": "\n",
    "cr": "\r",
    "crlf": "\r\n",
}

FIELDS = {
    "{comma}": "A comma: ','",
    "{semicolon}": "A semicolon: ';'",
    "{questionmark}": "A question mark: '?'",
    "{pipe}": "A vertical pipe: '|'",
    "{percent}": "A percent sign: '%'",
    "{openbrace}": "An open brace: '{'",
    "{closebrace}": "A close brace: '}'",
    "{openparens}": "An open parentheses: '('",
    "{closeparens}": "A close parentheses: ')'",
    "{openbracket}": "An open bracket: '['",
    "{closebracket}": "A close bracket: ']'",
    "{newline}": r"A newline: '\n'",
    "{lf}": r"A line feed: '\n', alias for {newline}",
    "{cr}": r"A carriage return: '\r'",
    "{crlf}": r"a carriage return + line feed: '\r\n'",
}


@autofile.hookimpl
def get_template_help() -> Iterable:
    text = """
    Within the template system, many punctuation characters have special meaning, e.g. `{}` indicates a template field 
    and this means that some punctuation characters cannot be inserted into the template. 
    Thus, if you want to insert punctuation into the rendered template value, you can use these punctuation fields to do so. 
    For example, `{openbrace}value{closebrace}` will render to `{value}`.
    """
    fields = [["Field", "Description"], *[[k, v] for k, v in FIELDS.items()]]
    return ["**Punctuation Fields**", fields, text]


@autofile.hookimpl
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """Called by template.py to get template value for custom template"""
    value = PUNCTUATION.get(field)
    return [value] if value else None



"""Template plugins for autofile to format strings and strip whitespace"""

from typing import Iterable, List, Optional

import autofile
from autofile.datetime_formatter import DateTimeFormatter

FIELDS = {
    "{strip}": "Use in form '{strip,TEMPLATE}'; strips whitespace from begining and end of rendered TEMPLATE value(s).",
    "{format}": "Use in form, '{format:TYPE:FORMAT,TEMPLATE}'; converts TEMPLATE value to TYPE then formats the value using python string formatting codes specified by FORMAT; TYPE is one of: 'int', 'float', or 'str'.",
}


@autofile.hookimpl
def get_template_help() -> Iterable:
    text = """
    The `{strip}` and `{format}` fields are used to format strings. 
    `{strip,TEMPLATE}` strips whitespace from TEMPLATE. 
    For example, `{strip,{exiftool:Title}}` will strip any excess whitespace from the title of an image file. 

    `{format:TYPE:FORMAT,TEMPLATE}` formats TEMPLATE using python string formatting codes. 
    For example: 
    
    - `{format:int:02d,{audio:track}}` will format the track number of an audio file to two digits with leading zeros. 
    - `{format:str:-^30,{audio.title}}` will center the title of an audio file and pad it to 30 characters with '-'.

    TYPE must be one of 'int', 'float', or 'str'. 
    See https://docs.python.org/3.7/library/string.html#formatspec for more information on valid FORMAT values.
    """
    fields = [["Field", "Description"], *[[k, v] for k, v in FIELDS.items()]]
    return ["**String Formatting Fields**", fields, text]


@autofile.hookimpl
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """lookup value for file dates

    Args:
        field: template field to find value for.

    Returns:
        The matching template value (which may be None).
    """
    if field == "strip":
        return [v.strip() for v in default]

    if field == "format":
        if not subfield or ":" not in subfield:
            raise ValueError("{format} requires subfield in form TYPE:FORMAT")
        type_, format_str = subfield.split(":", 1)
        if type_ not in ("int", "float", "str"):
            raise ValueError(
                f"'{type_}' is not a valid type for {format}: must be one of 'int', 'float', 'str'"
            )
        if type_ == "int":
            default = [int(v) for v in default]
        elif type_ == "float":
            default = [float(v) for v in default]
        return [format_str_value(v, format_str) for v in default]

    return None


def format_str_value(value, format_str):
    """Format value based on format code in field in format id:02d"""
    if not format_str:
        return str(value)
    format_str = "{0:" + f"{format_str}" + "}"
    return format_str.format(value)



"""template plugin for audio files for autofile"""

""" Works on the following audio formats:
    Wave/RIFF
    OGG
    OPUS
    FLAC
    WMA
    MP4/M4A/M4B
    AIFF/AIFF-C
"""

from typing import Iterable, List, Optional

from tinytag import TinyTag, TinyTagException

import autofile

FIELDS = {
    "{audio}": "Use in form '{audio:TAG}'; Returns tag value for various audio types include mp3, "
}

SUBFIELDS = {
    "album": "album as string",
    "albumartist": "album artist as string",
    "artist": "artist name as string",
    "audio_offset": "number of bytes before audio data begins",
    "bitrate": "bitrate in kBits/s",
    "comment": "file comment as string",
    "composer": "composer as string",
    "disc": "disc number",
    "disc_total": "the total number of discs",
    "duration": "duration of the song in seconds",
    "filesize": "file size in bytes",
    "genre": "genre as string",
    "samplerate": "samples per second",
    "title": "title of the song",
    "track": "track number as string",
    "track_total": "total number of tracks as string",
    "year": "year or data as string",
}


@autofile.hookimpl
def get_template_help() -> Iterable:
    text = """
    The `{audio}` field provides access to audio-file related tags for audio files. The following formats are supported:

    - MP3 (ID3 v1, v1.1, v2.2, v2.3+)
    - Wave/RIFF
    - OGG
    - OPUS
    - FLAC
    - WMA
    - MP4/M4A/M4B
    - AIFF/AIFF-C

    The `{audio}` field must be used with one or more the following subfields in the form: `{audio:SUBFIELD}`, 
    for example: `{audio:title}` or `{audio:artist}`. 
    """

    fields = [["Field", "Description"], *[[k, v] for k, v in FIELDS.items()]]
    subfields = [["Subfield", "Description"], *[[k, v] for k, v in SUBFIELDS.items()]]
    return ["**Audio Files**", fields, text, subfields]


@autofile.hookimpl
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """lookup value for file dates

    Args:
        field: template field to find value for.

    Returns:
        The matching template value (which may be None).
    """
    if "{" + field + "}" not in FIELDS:
        return None

    try:
        tag = TinyTag.get(filepath)
        vals = []
        if field == "audio":
            if subfield is None:
                raise ValueError("subfield must be specified for audio field")
            if subfield not in SUBFIELDS:
                raise ValueError(f"Unknown audio subfield: {subfield}")
            vals = str(getattr(tag, subfield))
        return [vals]
    except TinyTagException as e:
        autofile.cli.print_warning(
            f"Error reading tag {field}:{subfield} for file {filepath}: {e}"
        )
        return [None]



"""Plugin for autofile template to process file dates"""

import datetime
import os
from typing import Iterable, List, Optional

import autofile
from autofile.datetime_formatter import DateTimeFormatter
from autofile.datetime_utils import datetime_utc_to_local

TODAY = None

FIELDS = {
    "{created}": "File creation date/time",
    "{modified}": "File modification date/time",
    "{accessed}": "File last accessed date/time",
    "{today}": "The current date/time (as of when {today} is first evaluated)",
    "{now}": "The current date/time (evaluated at the time the template is processed)",
}

DATETIME_ATTRIBUTES = {
    "date": "ISO date, e.g. 2020-03-22",
    "year": "4-digit year, e.g. 2021",
    "yy": "2-digit year, e.g. 21",
    "month": "Month name as locale's full name, e.g. December",
    "mon": "Month as locale's abbreviated name, e.g. Dec",
    "mm": "2-digit month, e.g. 12",
    "dd": "2-digit day of the month, e.g. 22",
    "dow": "Day of the week as locale's full name, e.g. Tuesday",
    "doy": "Julian day of year starting from 001",
    "hour": "2-digit hour, e.g. 10",
    "min": "2-digit minute, e.g. 15",
    "sec": "2-digit second, e.g. 30",
    "strftime": "Apply strftime template to date/time. Should be used in form "
    + "{created.strftime,TEMPLATE} where TEMPLATE is a valid strftime template, e.g. "
    + "{created.strftime,%Y-%U} would result in year-week number of year: '2020-23'. "
    + "If used with no template will return null value. "
    + "See https://strftime.org/ for help on strftime templates.",
}


@autofile.hookimpl
def get_template_help() -> Iterable:
    text = """
    Date/time fields may be formatted using "dot notation" attributes which are appended to the field name following a `.` (period). 
    For example, `{created.month}` resolves to the month name of the file's creation date in the user's locale, e.g. `December`. 

    The `{today}` and `{now}` fields resolve to the current date/time with one key distinction between them: 
    `{today}` is the current date/time as of when `{today}` is first evaluated and will remain unchanged for every file processed; 
    `{now}` is the current date/time at the time each template is processed and will change with every file processed. 

    The following attributes are available:
    
    """
    fields = [["Field", "Description"], *[[k, v] for k, v in FIELDS.items()]]
    attributes = [
        ["Attribute", "Description"],
        *[[k, v] for k, v in DATETIME_ATTRIBUTES.items()],
    ]
    return ["**Date/Time Fields**", fields, text, attributes]


@autofile.hookimpl
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """lookup value for file dates

    Args:
        field: template field to find value for.

    Returns:
        The matching template value (which may be None).
    """
    global TODAY

    field = field.split(".", 1)
    if "{" + field[0] + "}" not in FIELDS:
        return None

    stat_info = os.stat(filepath)
    dt = None
    if field[0] == "created":
        dt = datetime.datetime.fromtimestamp(stat_info.st_birthtime)
    elif field[0] == "modified":
        dt = datetime.datetime.fromtimestamp(stat_info.st_mtime)
    elif field[0] == "accessed":
        dt = datetime.datetime.fromtimestamp(stat_info.st_atime)
    elif field[0] == "today":
        if TODAY is None:
            # initialize TODAY only the first time the template is evaluated
            TODAY = datetime_utc_to_local(datetime.datetime.now(datetime.timezone.utc))
        dt = TODAY
    elif field[0] == "now":
        dt = datetime_utc_to_local(datetime.datetime.now(datetime.timezone.utc))
    else:
        raise ValueError(f"Unknown field {field}")

    if len(field) == 1:
        return [dt.isoformat()]

    subfield = field[1]
    if subfield not in DATETIME_ATTRIBUTES:
        raise ValueError(f"Unknown subfield {subfield}")

    if subfield == "strftime":
        if default:
            try:
                value = dt.strftime(default[0]) if dt else None
            except:
                raise ValueError(f"Invalid strftime template: '{default}'")
        else:
            value = None
        return [value]
    return [getattr(DateTimeFormatter(dt), subfield)]



"""Utils for autofile"""

import platform


def pluralize(count, singular, plural):
    """Return singular or plural based on count"""
    if count == 1:
        return singular
    else:
        return plural


def noop(*args, **kwargs):
    """No-op function for use as verbose if verbose not set"""
    pass


def red(msg: str) -> str:
    """Return red string in rich markup"""
    return f"[red]{msg}[/red]"


def green(msg: str) -> str:
    """Return green string in rich markup"""
    return f"[green]{msg}[/green]"


def bold(msg: str) -> str:
    """Return bold string in rich markup"""
    return f"[bold]{msg}[/bold]"


def get_os_version():
    # returns tuple of str containing OS version
    # e.g. 10.13.6 = ("10", "13", "6")
    version = platform.mac_ver()[0].split(".")
    if len(version) == 2:
        (ver, major) = version
        minor = "0"
    elif len(version) == 3:
        (ver, major, minor) = version
    else:
        raise (
            ValueError(
                f"Could not parse version string: {platform.mac_ver()} {version}"
            )
        )
    return (ver, major, minor)



"""plugin for autofile to access Finder metadata (e.g. tags, comments)"""

from osxmetadata import OSXMetaData
from typing import Dict, Iterable, List, Optional

import autofile
from autofile.datetime_utils import datetime_naive_to_utc

FIELDS = {
    "{finder}": "Get metadata managed by macOS Finder such as tags and comments; use in form '{finder:SUBFIELD}', e.g. '{finder:tags}'",
}

SUBFIELDS = {"tags": "Finder tags (keywords)", "comment": "Finder comment"}


@autofile.hookimpl
def get_template_help() -> Iterable:
    text = """
    `{finder}` provides access to Finder metadata. It must be used in the form `{finder:SUBFIELD}`
    where SUBFIELD is one of the following:
    """

    fields = [["Field", "Description"], *[[k, v] for k, v in FIELDS.items()]]
    subfields = [["Subfield", "Description"], *[[k, v] for k, v in SUBFIELDS.items()]]
    return ["**Finder Metadata**", fields, text, subfields]


@autofile.hookimpl
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """lookup value for file dates

    Args:
        field: template field to find value for.

    Returns:
        The matching template value (which may be None).
    """
    if "{" + field + "}" not in FIELDS:
        return None
    if not subfield or subfield not in SUBFIELDS:
        raise ValueError(f"Missing or invalid subfield '{subfield}'")

    metadata = OSXMetaData(filepath)
    if subfield == "tags":
        tags = metadata.tags
        return [tag.name for tag in tags]
    elif subfield == "comment":
        comment = metadata.findercomment
        return [str(comment)]



"""Plugin for autofile template to process uti template """

""" Implementation note: runs only on macOS

    On macOS <= 11 (Big Sur), uses objective C CoreServices methods 
    UTTypeCopyPreferredTagWithClass and UTTypeCreatePreferredIdentifierForTag to retrieve the 
    UTI and the extension.  These are deprecated in 10.15 (Catalina) and no longer supported on Monterey.

    On Monterey, these calls are replaced with Swift methods that I can't call from python so 
    this code uses a cached dict of UTI values.  The code first checks to see if the extension or UTI 
    is available in the cache and if so, returns it. If not, it performs a subprocess call to `mdls` to 
    retrieve the UTI (by creating a temp file with the correct extension) and returns the UTI.  This only 
    works for the extension -> UTI lookup. On Monterey, if there is no cached value for UTI -> extension lookup, 
    returns None.

    It's a bit hacky but best I can think of to make this robust on different versions of macOS.  PRs welcome.
"""

import csv
import pathlib
import re
import subprocess
import tempfile
from typing import Iterable, List, Optional

import CoreServices
import objc

import autofile
from autofile.utils import get_os_version

FIELDS = {
    "{uti}": "Uniform Type Identifier (UTI) for file",
}


@autofile.hookimpl
def get_template_help() -> Iterable:
    text = """
    The `{uti}` template returns the macOS Uniform Type Identifier (UTI) for the file. 
    For example, `{uti}` for a file with extension .pdf would return `com.adobe.pdf`.
    """
    fields = [["Field", "Description"], *[[k, v] for k, v in FIELDS.items()]]
    return ["**Uniform Type Identifier (UTI) Fields**", fields, text]


@autofile.hookimpl
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """lookup value for file dates

    Args:
        field: template field to find value for.

    Returns:
        The matching template value (which may be None).
    """
    if field == "uti":
        return [get_uti_for_extension(pathlib.Path(filepath).suffix)]
    return None


# cached values of all the UTIs (< 6 chars long) known to my Mac running macOS 10.15.7
UTI_CSV = """extension,UTI,preferred_extension,MIME_type
c,public.c-source,c,None
f,public.fortran-source,f,None
h,public.c-header,h,None
i,public.c-source.preprocessed,i,None
l,public.lex-source,l,None
m,public.objective-c-source,m,None
o,public.object-code,o,None
r,com.apple.rez-source,r,None
s,public.assembly-source,s,None
y,public.yacc-source,y,None
z,public.z-archive,z,application/x-compress
aa,com.audible.aa-audiobook,aa,audio/audible
ai,com.adobe.illustrator.ai-image,ai,None
as,com.apple.applesingle-archive,as,None
au,public.au-audio,au,audio/basic
bz,public.bzip2-archive,bz2,application/x-bzip2
cc,public.c-plus-plus-source,cp,None
cp,public.c-plus-plus-source,cp,None
dv,public.dv-movie,dv,video/x-dv
gz,org.gnu.gnu-zip-archive,gz,application/x-gzip
hh,public.c-plus-plus-header,hh,None
hp,public.c-plus-plus-header,hh,None
ii,public.c-plus-plus-source.preprocessed,ii,None
js,com.netscape.javascript-source,js,text/javascript
lm,public.lex-source,l,None
mi,public.objective-c-source.preprocessed,mi,None
mm,public.objective-c-plus-plus-source,mm,None
pf,com.apple.colorsync-profile,icc,None
pl,public.perl-script,pl,text/x-perl-script
pm,public.perl-script,pl,text/x-perl-script
ps,com.adobe.postscript,ps,application/postscript
py,public.python-script,py,text/x-python-script
qt,com.apple.quicktime-movie,mov,video/quicktime
ra,com.real.realaudio,ram,audio/vnd.rn-realaudio
rb,public.ruby-script,rb,text/x-ruby-script
rm,com.real.realmedia,rm,application/vnd.rn-realmedia
sh,public.shell-script,sh,None
ts,public.mpeg-2-transport-stream,ts,None
ul,public.ulaw-audio,ul,None
uu,public.uuencoded-archive,uu,text/x-uuencode
wm,com.microsoft.windows-media-wm,wm,video/x-ms-wm
ym,public.yacc-source,y,None
aac,public.aac-audio,aac,audio/aac
aae,com.apple.photos.apple-adjustment-envelope,aae,None
aaf,org.aafassociation.advanced-authoring-format,aaf,None
aax,com.audible.aax-audiobook,aax,audio/vnd.audible.aax
abc,public.alembic,abc,None
ac3,public.ac3-audio,ac3,audio/ac3
ada,public.ada-source,ada,None
adb,public.ada-source,ada,None
ads,public.ada-source,ada,None
aif,public.aifc-audio,aifc,audio/aiff
amr,org.3gpp.adaptive-multi-rate-audio,amr,audio/amr
app,com.apple.application-bundle,app,None
arw,com.sony.arw-raw-image,arw,None
asf,com.microsoft.advanced-systems-format,asf,video/x-ms-asf
asx,com.microsoft.advanced-stream-redirector,asx,video/x-ms-asx
avi,public.avi,avi,video/avi
bdm,public.avchd-content,bdm,None
bin,com.apple.macbinary-archive,bin,application/macbinary
bmp,com.microsoft.bmp,bmp,image/bmp
bwf,com.microsoft.waveform-audio,wav,audio/vnd.wave
bz2,public.bzip2-archive,bz2,application/x-bzip2
caf,com.apple.coreaudio-format,caf,None
cdr,com.apple.disk-image-cdr,dvdr,None
cel,public.flc-animation,flc,video/flc
cer,public.x509-certificate,cer,application/x-x509-ca-cert
cpp,public.c-plus-plus-source,cp,None
crt,public.x509-certificate,cer,application/x-x509-ca-cert
crw,com.canon.crw-raw-image,crw,image/x-canon-crw
cr2,com.canon.cr2-raw-image,cr2,None
cr3,com.canon.cr3-raw-image,cr3,None
csh,public.csh-script,csh,None
css,public.css,css,text/css
csv,public.comma-separated-values-text,csv,text/csv
cxx,public.c-plus-plus-source,cp,None
dae,org.khronos.collada.digital-asset-exchange,dae,None
dcm,org.nema.dicom,dcm,application/dicom
dcr,com.kodak.raw-image,dcr,None
dds,com.microsoft.dds,dds,None
der,public.x509-certificate,cer,application/x-x509-ca-cert
dif,public.dv-movie,dv,video/x-dv
dll,com.microsoft.windows-dynamic-link-library,dll,application/x-msdownload
dls,public.downloadable-sound,dls,audio/dls
dmg,com.apple.disk-image-udif,dmg,None
dng,com.adobe.raw-image,dng,image/x-adobe-dng
doc,com.microsoft.word.doc,doc,application/msword
dot,com.microsoft.word.dot,dot,application/msword
dxo,com.dxo.raw-image,dxo,image/x-dxo-dxo
ec3,public.enhanced-ac3-audio,eac3,audio/eac3
edn,com.adobe.edn,edn,None
efx,com.j2.efx-fax,efx,image/efax
eml,com.apple.mail.email,eml,message/rfc822
eps,com.adobe.encapsulated-postscript,eps,None
erf,com.epson.raw-image,erf,image/x-epson-erf
etd,com.adobe.etd,etd,None
exe,com.microsoft.windows-executable,exe,application/x-msdownload
exp,com.apple.symbol-export,exp,None
exr,com.ilm.openexr-image,exr,None
fdf,com.adobe.fdf,fdf,None
fff,com.hasselblad.fff-raw-image,fff,None
flc,public.flc-animation,flc,video/flc
fli,public.flc-animation,flc,video/flc
flv,com.adobe.flash.video,flv,video/x-flv
for,public.fortran-source,f,None
fpx,com.kodak.flashpix-image,fpx,image/fpx
f4a,com.adobe.flash.video,flv,video/x-flv
f4b,com.adobe.flash.video,flv,video/x-flv
f4p,com.adobe.flash.video,flv,video/x-flv
f4v,com.adobe.flash.video,flv,video/x-flv
f77,public.fortran-77-source,f77,None
f90,public.fortran-90-source,f90,None
f95,public.fortran-95-source,f95,None
gif,com.compuserve.gif,gif,image/gif
hdr,public.radiance,pic,None
hpp,public.c-plus-plus-header,hh,None
hqx,com.apple.binhex-archive,hqx,application/mac-binhex40
htm,public.html,html,text/html
hxx,public.c-plus-plus-header,hh,None
iba,com.apple.ibooksauthor.pkgbook,iba,None
icc,com.apple.colorsync-profile,icc,None
icm,com.apple.colorsync-profile,icc,None
ico,com.microsoft.ico,ico,image/vnd.microsoft.icon
ics,com.apple.ical.ics,ics,text/calendar
iig,com.apple.iig-source,iig,None
iiq,com.phaseone.raw-image,iiq,None
img,com.apple.disk-image-ndif,ndif,None
inl,public.c-plus-plus-inline-header,inl,None
ipa,com.apple.itunes.ipa,ipa,None
ipp,public.c-plus-plus-header,hh,None
ips,com.apple.ips,ips,None
iso,public.iso-image,iso,None
ite,com.apple.tv.ite,ite,None
itl,com.apple.itunes.db,itl,None
jar,com.sun.java-archive,jar,application/java-archive
jav,com.sun.java-source,java,None
jfx,com.j2.jfx-fax,jfx,None
jpe,public.jpeg,jpeg,image/jpeg
jpf,public.jpeg-2000,jp2,image/jp2
jpg,public.jpeg,jpeg,image/jpeg
jpx,public.jpeg-2000,jp2,image/jp2
jp2,public.jpeg-2000,jp2,image/jp2
j2c,public.jpeg-2000,jp2,image/jp2
j2k,public.jpeg-2000,jp2,image/jp2
kar,public.midi-audio,midi,audio/midi
key,com.apple.iwork.keynote.key,key,None
ksh,public.ksh-script,ksh,None
kth,com.apple.iwork.keynote.kth,kth,None
ktx,org.khronos.ktx,ktx,None
lid,public.dylan-source,dlyan,None
lmm,public.lex-source,l,None
log,com.apple.log,log,None
lpp,public.lex-source,l,None
lxx,public.lex-source,l,None
mid,public.midi-audio,midi,audio/midi
mig,public.mig-source,defs,None
mii,public.objective-c-plus-plus-source.preprocessed,mii,None
mjs,com.netscape.javascript-source,js,text/javascript
mnc,ca.mcgill.mni.bic.mnc,mnc,None
mos,com.leafamerica.raw-image,mos,None
mov,com.apple.quicktime-movie,mov,video/quicktime
mpe,public.mpeg,mpg,video/mpeg
mpg,public.mpeg,mpg,video/mpeg
mpo,public.mpo-image,mpo,None
mp2,public.mp2,mp2,None
mp3,public.mp3,mp3,audio/mpeg
mp4,public.mpeg-4,mp4,video/mp4
mrw,com.konicaminolta.raw-image,mrw,None
mts,public.avchd-mpeg-2-transport-stream,mts,None
mxf,org.smpte.mxf,mxf,application/mxf
m15,public.mpeg,mpg,video/mpeg
m2v,public.mpeg-2-video,m2v,video/mpeg2
m3u,public.m3u-playlist,m3u,audio/mpegurl
m4a,com.apple.m4a-audio,m4a,audio/x-m4a
m4b,com.apple.protected-mpeg-4-audio-b,m4b,None
m4p,com.apple.protected-mpeg-4-audio,m4p,None
m4r,com.apple.mpeg-4-ringtone,m4r,audio/x-m4r
m4v,com.apple.m4v-video,m4v,video/x-m4v
m75,public.mpeg,mpg,video/mpeg
nef,com.nikon.raw-image,nef,None
nii,gov.nih.nifti-1,nii,None
nrw,com.nikon.nrw-raw-image,nrw,image/x-nikon-nrw
obj,public.geometry-definition-format,obj,None
odb,org.oasis-open.opendocument.database,odb,application/vnd.oasis.opendocument.database
odc,org.oasis-open.opendocument.chart,odc,application/vnd.oasis.opendocument.chart
odf,org.oasis-open.opendocument.formula,odf,application/vnd.oasis.opendocument.formula
odg,org.oasis-open.opendocument.graphics,odg,application/vnd.oasis.opendocument.graphics
odi,org.oasis-open.opendocument.image,odi,application/vnd.oasis.opendocument.image
odm,org.oasis-open.opendocument.text-master,odm,application/vnd.oasis.opendocument.text-master
odp,org.oasis-open.opendocument.presentation,odp,application/vnd.oasis.opendocument.presentation
ods,org.oasis-open.opendocument.spreadsheet,ods,application/vnd.oasis.opendocument.spreadsheet
odt,org.oasis-open.opendocument.text,odt,application/vnd.oasis.opendocument.text
omf,com.avid.open-media-framework,omf,None
orf,com.olympus.raw-image,orf,None
otc,public.opentype-collection-font,otc,None
otf,public.opentype-font,otf,None
otg,org.oasis-open.opendocument.graphics-template,otg,application/vnd.oasis.opendocument.graphics-template
oth,org.oasis-open.opendocument.text-web,oth,application/vnd.oasis.opendocument.text-web
oti,org.oasis-open.opendocument.image-template,oti,application/vnd.oasis.opendocument.image-template
otp,org.oasis-open.opendocument.presentation-template,otp,application/vnd.oasis.opendocument.presentation-template
ots,org.oasis-open.opendocument.spreadsheet-template,ots,application/vnd.oasis.opendocument.spreadsheet-template
ott,org.oasis-open.opendocument.text-template,ott,application/vnd.oasis.opendocument.text-template
pas,public.pascal-source,pas,None
pax,public.cpio-archive,cpio,None
pbm,public.pbm,pbm,None
pch,public.precompiled-c-header,pch,None
pct,com.apple.pict,pict,image/pict
pdf,com.adobe.pdf,pdf,application/pdf
pef,com.pentax.raw-image,pef,None
pem,public.x509-certificate,cer,application/x-x509-ca-cert
pfa,com.adobe.postscript-pfa-font,pfa,None
pfb,com.adobe.postscript-pfb-font,pfb,None
pfm,public.pbm,pbm,None
pfx,com.rsa.pkcs-12,p12,application/x-pkcs12
pgm,public.pbm,pbm,None
pgn,com.apple.chess.pgn,pgn,None
php,public.php-script,php,text/php
ph3,public.php-script,php,text/php
ph4,public.php-script,php,text/php
pic,com.apple.pict,pict,image/pict
pkg,com.apple.installer-package-archive,pkg,None
pls,public.pls-playlist,pls,audio/x-scpls
ply,public.polygon-file-format,ply,None
png,public.png,png,image/png
pot,com.microsoft.powerpoint.pot,pot,application/vnd.ms-powerpoint
ppm,public.pbm,pbm,None
pps,com.microsoft.powerpoint.pps,pps,application/vnd.ms-powerpoint
ppt,com.microsoft.powerpoint.ppt,ppt,application/vnd.ms-powerpoint
psb,com.adobe.photoshop-large-image,psb,None
psd,com.adobe.photoshop-image,psd,image/vnd.adobe.photoshop
pvr,public.pvr,pvr,None
pvt,com.apple.private.live-photo-bundle,pvt,None
pwl,com.leica.pwl-raw-image,pwl,image/x-leica-pwl
p12,com.rsa.pkcs-12,p12,application/x-pkcs12
qti,com.apple.quicktime-image,qtif,image/x-quicktime
qtz,com.apple.quartz-composer-composition,qtz,application/x-quartzcomposer
raf,com.fuji.raw-image,raf,None
ram,com.real.realaudio,ram,audio/vnd.rn-realaudio
raw,com.panasonic.raw-image,raw,None
rbw,public.ruby-script,rb,text/x-ruby-script
rmp,com.apple.music.rmp-playlist,rmp,application/vnd.rn-rn_music_package
rss,public.rss,rss,application/rss+xml
rtf,public.rtf,rtf,text/rtf
rwl,com.leica.rwl-raw-image,rwl,None
rw2,com.panasonic.rw2-raw-image,rw2,None
scc,com.scenarist.closed-caption,scc,None
scn,com.apple.scenekit.scene,scn,None
sda,org.openoffice.graphics,sxd,application/vnd.sun.xml.draw
sdc,org.openoffice.spreadsheet,sxc,application/vnd.sun.xml.calc
sdd,org.openoffice.presentation,sxi,application/vnd.sun.xml.impress
sdp,org.openoffice.presentation,sxi,application/vnd.sun.xml.impress
sdv,public.3gpp,3gp,video/3gpp
sdw,org.openoffice.text,sxw,application/vnd.sun.xml.writer
sd2,com.digidesign.sd2-audio,sd2,None
sea,com.stuffit.archive.sit,sit,application/x-stuffit
sf2,com.soundblaster.soundfont,sf2,None
sgi,com.sgi.sgi-image,sgi,image/sgi
sit,com.stuffit.archive.sit,sit,application/x-stuffit
slm,com.apple.photos.slow-motion-video-sidecar,slm,None
smf,public.midi-audio,midi,audio/midi
smi,com.apple.disk-image-smi,smi,None
snd,public.au-audio,au,audio/basic
spx,com.apple.systemprofiler.document,spx,None
srf,com.sony.raw-image,srf,None
srw,com.samsung.raw-image,srw,None
sr2,com.sony.sr2-raw-image,sr2,image/x-sony-sr2
stc,org.openoffice.spreadsheet-template,stc,application/vnd.sun.xml.calc.template
std,org.openoffice.graphics-template,std,application/vnd.sun.xml.draw.template
sti,org.openoffice.presentation-template,sti,application/vnd.sun.xml.impress.template
stl,public.standard-tesselated-geometry-format,stl,None
stw,org.openoffice.text-template,stw,application/vnd.sun.xml.writer.template
svg,public.svg-image,svg,image/svg+xml
sxc,org.openoffice.spreadsheet,sxc,application/vnd.sun.xml.calc
sxd,org.openoffice.graphics,sxd,application/vnd.sun.xml.draw
sxg,org.openoffice.text-master,sxg,application/vnd.sun.xml.writer.global
sxi,org.openoffice.presentation,sxi,application/vnd.sun.xml.impress
sxm,org.openoffice.formula,sxm,application/vnd.sun.xml.math
sxw,org.openoffice.text,sxw,application/vnd.sun.xml.writer
tar,public.tar-archive,tar,application/x-tar
tbz,public.tar-bzip2-archive,tbz2,None
tga,com.truevision.tga-image,tga,image/targa
tgz,org.gnu.gnu-zip-tar-archive,tgz,None
tif,public.tiff,tiff,image/tiff
tsv,public.tab-separated-values-text,tsv,text/tab-separated-values
ttc,public.truetype-collection-font,ttc,None
ttf,public.truetype-ttf-font,ttf,None
txt,public.plain-text,txt,text/plain
ulw,public.ulaw-audio,ul,None
url,com.microsoft.internet-shortcut,url,None
usd,com.pixar.universal-scene-description,usd,None
vcf,public.vcard,vcf,text/directory
vcs,com.apple.ical.vcs,vcs,text/x-vcalendar
vfw,public.avi,avi,video/avi
vtt,org.w3.webvtt,vtt,text/vtt
war,com.sun.web-application-archive,war,None
wav,com.microsoft.waveform-audio,wav,audio/vnd.wave
wax,com.microsoft.windows-media-wax,wax,video/x-ms-wax
web,com.getdropbox.dropbox.shortcut,web,None
wma,com.microsoft.windows-media-wma,wma,video/x-ms-wma
wmp,com.microsoft.windows-media-wmp,wmp,video/x-ms-wmp
wmv,com.microsoft.windows-media-wmv,wmv,video/x-ms-wmv
wmx,com.microsoft.windows-media-wmx,wmx,video/x-ms-wmx
wvx,com.microsoft.windows-media-wvx,wvx,video/x-ms-wvx
xar,com.apple.xar-archive,xar,None
xbm,public.xbitmap-image,xbm,image/x-xbitmap
xfd,public.xfd,xfd,None
xht,public.xhtml,xhtml,application/xhtml+xml
xip,com.apple.xip-archive,xip,None
xla,com.microsoft.excel.xla,xla,None
xls,com.microsoft.excel.xls,xls,application/vnd.ms-excel
xlt,com.microsoft.excel.xlt,xlt,application/vnd.ms-excel
xlw,com.microsoft.excel.xlw,xlw,application/vnd.ms-excel
xml,public.xml,xml,application/xml
xmp,com.seriflabs.xmp,xmp,application/rdf+xml
xpc,com.apple.xpc-service,xpc,None
yml,public.yaml,yml,application/x-yaml
ymm,public.yacc-source,y,None
ypp,public.yacc-source,y,None
yxx,public.yacc-source,y,None
zip,public.zip-archive,zip,application/zip
zsh,public.zsh-script,zsh,None
3fr,com.hasselblad.3fr-raw-image,3fr,None
3gp,public.3gpp,3gp,video/3gpp
3g2,public.3gpp2,3g2,video/3gpp2
adts,public.aac-audio,aac,audio/aac
ahap,com.apple.haptics.ahap,ahap,None
aifc,public.aifc-audio,aifc,audio/aiff
aiff,public.aifc-audio,aifc,audio/aiff
astc,org.khronos.astc,astc,None
avci,public.avci,avci,image/avci
avcs,public.avcs,avcs,image/avcs
band,com.apple.garageband.project,band,None
bash,public.bash-script,bash,None
bdmv,public.avchd-content,bdm,None
book,com.apple.ibooksauthor.pkgbook,iba,None
cdda,public.aifc-audio,aifc,audio/aiff
chat,com.apple.ichat.transcript,ichat,None
cpgz,com.apple.bom-compressed-cpio,cpgz,None
cpio,public.cpio-archive,cpio,None
dart,com.apple.disk-image-dart,dart,None
dc42,com.apple.disk-image-dc42,dc42,None
defs,public.mig-source,defs,None
dext,com.apple.driver-extension,dext,None
diff,public.patch-file,patch,None
dist,com.apple.installer-distribution-package,dist,None
docm,org.openxmlformats.wordprocessingml.document.macroenabled,docm,application/vnd.ms-word.document.macroenabled.12
docx,org.openxmlformats.wordprocessingml.document,docx,application/vnd.openxmlformats-officedocument.wordprocessingml.document
dotm,org.openxmlformats.wordprocessingml.template.macroenabled,dotm,application/vnd.ms-word.template.macroenabled.12
dotx,org.openxmlformats.wordprocessingml.template,dotx,application/vnd.openxmlformats-officedocument.wordprocessingml.template
dsym,com.apple.xcode.dsym,dsym,None
dvdr,com.apple.disk-image-cdr,dvdr,None
eac3,public.enhanced-ac3-audio,eac3,audio/eac3
emlx,com.apple.mail.emlx,emlx,None
enex,com.evernote.enex,enex,None
epub,org.idpf.epub-container,epub,application/epub+zip
fh10,com.seriflabs.affinity,fh10,None
fh11,com.seriflabs.affinity,fh10,None
flac,org.xiph.flac,flac,audio/flac
fpbf,com.apple.finder.burn-folder,fpbf,None
game,com.apple.chess.game,game,None
gdoc,com.google.gdoc,gdoc,None
gtar,org.gnu.gnu-tar-archive,gtar,application/x-gtar
gzip,org.gnu.gnu-zip-archive,gz,application/x-gzip
hang,com.apple.hangreport,hang,None
heic,public.heic,heic,image/heic
heif,public.heif,heif,image/heif
html,public.html,html,text/html
hvpl,com.apple.music.visual,hvpl,None
icbu,com.apple.ical.backup,icbu,None
icns,com.apple.icns,icns,None
ipsw,com.apple.itunes.ipsw,ipsw,None
itlp,com.apple.music.itlp,itlp,None
itms,com.apple.itunes.store-url,itms,None
java,com.sun.java-source,java,None
jnlp,com.sun.java-web-start,jnlp,application/x-java-jnlp-file
jpeg,public.jpeg,jpeg,image/jpeg
json,public.json,json,application/json
latm,public.mp4a-loas,loas,None
loas,public.mp4a-loas,loas,None
lpdf,com.apple.localized-pdf-bundle,lpdf,None
mbox,com.apple.mail.mbox,mbox,None
menu,com.apple.menu-extra,menu,None
midi,public.midi-audio,midi,audio/midi
minc,ca.mcgill.mni.bic.mnc,mnc,None
mpeg,public.mpeg,mpg,video/mpeg
mpga,public.mp3,mp3,audio/mpeg
mpg4,public.mpeg-4,mp4,video/mp4
mpkg,com.apple.installer-package-archive,pkg,None
m2ts,public.avchd-mpeg-2-transport-stream,mts,None
m3u8,public.m3u-playlist,m3u,audio/mpegurl
ndif,com.apple.disk-image-ndif,ndif,None
note,com.apple.notes.note,note,None
php3,public.php-script,php,text/php
php4,public.php-script,php,text/php
pict,com.apple.pict,pict,image/pict
pntg,com.apple.macpaint-image,pntg,None
potm,org.openxmlformats.presentationml.template.macroenabled,potm,application/vnd.ms-powerpoint.template.macroenabled.12
potx,org.openxmlformats.presentationml.template,potx,application/vnd.openxmlformats-officedocument.presentationml.template
ppsm,org.openxmlformats.presentationml.slideshow.macroenabled,ppsm,application/vnd.ms-powerpoint.slideshow.macroenabled.12
ppsx,org.openxmlformats.presentationml.slideshow,ppsx,application/vnd.openxmlformats-officedocument.presentationml.slideshow
pptm,org.openxmlformats.presentationml.presentation.macroenabled,pptm,application/vnd.ms-powerpoint.presentation.macroenabled.12
pptx,org.openxmlformats.presentationml.presentation,pptx,application/vnd.openxmlformats-officedocument.presentationml.presentation
pset,com.apple.pdf-printer-settings,pset,None
qtif,com.apple.quicktime-image,qtif,image/x-quicktime
rmvb,com.real.realmedia-vbr,rmvb,application/vnd.rn-realmedia-vbr
rtfd,com.apple.rtfd,rtfd,None
scnz,com.apple.scenekit.scene,scn,None
scpt,com.apple.applescript.script,scpt,None
shtm,public.html,html,text/html
sidx,com.stuffit.archive.sidx,sidx,application/x-stuffitx-index
sitx,com.stuffit.archive.sitx,sitx,application/x-stuffitx
spin,com.apple.spinreport,spin,None
suit,com.apple.font-suitcase,suit,None
svgz,public.svg-image,svg,image/svg+xml
tbz2,public.tar-bzip2-archive,tbz2,None
tcsh,public.tcsh-script,tcsh,None
term,com.apple.terminal.session,term,None
text,public.plain-text,txt,text/plain
tiff,public.tiff,tiff,image/tiff
tool,com.apple.terminal.shell-script,command,None
udif,com.apple.disk-image-udif,dmg,None
ulaw,public.ulaw-audio,ul,None
usda,com.pixar.universal-scene-description,usd,None
usdc,com.pixar.universal-scene-description,usd,None
usdz,com.pixar.universal-scene-description-mobile,usdz,model/vnd.usdz+zip
vcal,com.apple.ical.vcs,vcs,text/x-vcalendar
wave,com.microsoft.waveform-audio,wav,audio/vnd.wave
wdgt,com.apple.dashboard-widget,wdgt,None
webp,public.webp,webp,None
xfdf,com.adobe.xfdf,xfdf,None
xhtm,public.xhtml,xhtml,application/xhtml+xml
xlsb,com.microsoft.excel.sheet.binary.macroenabled,xlsb,application/vnd.ms-excel.sheet.binary.macroenabled.12
xlsm,org.openxmlformats.spreadsheetml.sheet.macroenabled,xlsm,application/vnd.ms-excel.sheet.macroenabled.12
xlsx,org.openxmlformats.spreadsheetml.sheet,xlsx,application/vnd.openxmlformats-officedocument.spreadsheetml.sheet
xltm,org.openxmlformats.spreadsheetml.template.macroenabled,xltm,application/vnd.ms-excel.template.macroenabled.12
xltx,org.openxmlformats.spreadsheetml.template,xltx,application/vnd.openxmlformats-officedocument.spreadsheetml.template
yaml,public.yaml,yml,application/x-yaml
3gpp,public.3gpp,3gp,video/3gpp
3gp2,public.3gpp2,3g2,video/3gpp2
abcdg,com.apple.addressbook.group,abcdg,None
abcdp,com.apple.addressbook.person,abcdp,None
afpub,com.seriflabs.affinitypublisher.document,afpub,None
appex,com.apple.application-and-system-extension,appex,None
avchd,public.avchd-collection,avchd,None
blank,com.apple.preview.blank,blank,None
class,com.sun.java-class,class,None
crash,com.apple.crashreport,crash,None
dfont,com.apple.truetype-datafork-suitcase-font,dfont,None
dicom,org.nema.dicom,dcm,application/dicom
distz,com.apple.installer-distribution-package,dist,None
dlyan,public.dylan-source,dlyan,None
dylib,com.apple.mach-o-dylib,dylib,None
heics,public.heics,heics,image/heic-sequence
heifs,public.heifs,heifs,image/heif-sequence
ichat,com.apple.ichat.transcript,ichat,None
pages,com.apple.iwork.pages.pages,pages,None
panic,com.apple.panicreport,panic,None
paper,com.getdropbox.dropbox.paper,paper,None
patch,public.patch-file,patch,None
phtml,public.php-script,php,text/php
plist,com.apple.property-list,plist,None
saver,com.apple.systempreference.screen-saver,saver,None
scptd,com.apple.applescript.script-bundle,scptd,None
sfont,com.apple.cfr-font,sfont,None
shtml,public.html,html,text/html
swift,public.swift-source,swift,None
toast,com.roxio.disk-image-toast,toast,None
vcard,public.vcard,vcf,text/directory
wdmon,com.apple.wireless-diagnostics.wdmon,wdmon,None
xhtml,public.xhtml,xhtml,application/xhtml+xml
action,com.apple.automator-action,action,None
afploc,com.apple.afp-internet-location,afploc,None
"""

# load CSV separated uti data into dictionaries with key of extension and UTI
EXT_UTI_DICT = {}
UTI_EXT_DICT = {}


def _load_uti_dict():
    """load an initialize the cached UTI and extension dicts"""
    _reader = csv.DictReader(UTI_CSV.split("\n"), delimiter=",")
    for row in _reader:
        EXT_UTI_DICT[row["extension"]] = row["UTI"]
        UTI_EXT_DICT[row["UTI"]] = row["preferred_extension"]


_load_uti_dict()

# OS version for determining which methods can be used
OS_VER, OS_MAJOR, _ = (int(x) for x in get_os_version())


def _get_uti_from_mdls(extension):
    """use mdls to get the UTI for a given extension on systems that don't support UTTypeCreatePreferredIdentifierForTag
    Returns: UTI or None if UTI cannot be determined"""

    # mdls -name kMDItemContentType foo.3fr
    # kMDItemContentType = "com.hasselblad.3fr-raw-image"

    with tempfile.NamedTemporaryFile(suffix="." + extension) as temp:
        output = subprocess.check_output(
            [
                "/usr/bin/mdls",
                "-name",
                "kMDItemContentType",
                temp.name,
            ]
        ).splitlines()
        output = output[0].decode("UTF-8") if output else None
        if not output:
            return None

        match = re.match(r'kMDItemContentType\s+\=\s+"(.*)"', output)
        if match:
            return match.group(1)
        return None


def _get_uti_from_ext_dict(ext):
    try:
        return EXT_UTI_DICT[ext]
    except KeyError:
        return None


def _get_ext_from_uti_dict(uti):
    try:
        return UTI_EXT_DICT[uti]
    except KeyError:
        return None


def get_preferred_uti_extension(uti):
    """get preferred extension for a UTI type
    uti: UTI str, e.g. 'public.jpeg'
    returns: preferred extension as str or None if cannot be determined"""

    if (OS_VER, OS_MAJOR) <= (10, 16):
        # reference: https://developer.apple.com/documentation/coreservices/1442744-uttypecopypreferredtagwithclass?language=objc
        # deprecated in Catalina+, likely won't work at all on macOS 12
        with objc.autorelease_pool():
            extension = CoreServices.UTTypeCopyPreferredTagWithClass(
                uti, CoreServices.kUTTagClassFilenameExtension
            )
            if extension:
                return extension

            # on macOS 10.12, HEIC files are not supported and UTTypeCopyPreferredTagWithClass will return None for HEIC
            if uti == "public.heic":
                return "HEIC"

            return None

    return _get_ext_from_uti_dict(uti)


def get_uti_for_extension(extension):
    """get UTI for a given file extension"""

    # accepts extension with or without leading 0
    if extension[0] == ".":
        extension = extension[1:]

    if (OS_VER, OS_MAJOR) <= (10, 16):
        # https://developer.apple.com/documentation/coreservices/1448939-uttypecreatepreferredidentifierf
        with objc.autorelease_pool():
            uti = CoreServices.UTTypeCreatePreferredIdentifierForTag(
                CoreServices.kUTTagClassFilenameExtension, extension, None
            )
            if uti:
                return uti

            # on macOS 10.12, HEIC files are not supported and UTTypeCopyPreferredTagWithClass will return None for HEIC
            if extension.lower() == "heic":
                return "public.heic"

            return None

    uti = _get_uti_from_ext_dict(extension.lower())
    if uti:
        return uti

    uti = _get_uti_from_mdls(extension)
    if uti:
        # cache the UTI
        EXT_UTI_DICT[extension.lower()] = uti
        UTI_EXT_DICT[uti] = extension.lower()
        return uti

    return None



"""Plugin template for autofile to access Microsoft Word .docx file metadata"""

import datetime
import pathlib
from typing import Any, Iterable, List, Optional, Union

import docx

import autofile
from autofile.datetime_formatter import DateTimeFormatter

FIELDS = {
    "{docx}": "Access metadata properties of Microsoft Word document files (.docx); "
    "use in format {docx:SUBFIELD}"
}

SUBFIELDS = {
    "author": "Named ‘creator’ in spec. An entity primarily responsible for making the content of the resource. (Dublin Core)",
    "category": "A categorization of the content of this package. Example values for this property might include: Resume, Letter, Financial Forecast, Proposal, Technical Presentation, and so on. (Open Packaging Conventions)",
    "comments": "Named ‘description’ in spec. An explanation of the content of the resource. Values might include an abstract, table of contents, reference to a graphical representation of content, and a free-text account of the content. (Dublin Core)",
    "content_status": "The status of the content. Values might include “Draft”, “Reviewed”, and “Final”. (Open Packaging Conventions)",
    "created": "Date of creation of the resource; a date/time value. (Dublin Core)",
    "identifier": "An unambiguous reference to the resource within a given context. (Dublin Core)",
    "keywords": "A delimited set of keywords to support searching and indexing. This is typically a list of terms that are not available elsewhere in the properties. (Open Packaging Conventions)",
    "language": "The language of the intellectual content of the resource. (Dublin Core)",
    "last_modified_by": "The user who performed the last modification. The identification is environment-specific. Examples include a name, email address, or employee ID. It is recommended that this value be as concise as possible. (Open Packaging Conventions)",
    "last_printed": "The date and time of the last printing; a date/time value. (Open Packaging Conventions)",
    "modified": "Date on which the resource was changed; a date/time value. (Dublin Core)",
    "revision": "The revision number. This value might indicate the number of saves or revisions, provided the application updates it after each revision. (Open Packaging Conventions)",
    "subject": "The topic of the content of the resource. (Dublin Core)",
    "title": "The name given to the resource. (Dublin Core)",
    "version": "The version designator. This value is set by the user or by the application. (Open Packaging Conventions)",
}

DATETIME_SUBFIELDS = ["created", "modified", "last_printed"]

DATETIME_ATTRIBUTES = {
    "date": "ISO date, e.g. 2020-03-22",
    "year": "4-digit year, e.g. 2021",
    "yy": "2-digit year, e.g. 21",
    "month": "Month name as locale's full name, e.g. December",
    "mon": "Month as locale's abbreviated name, e.g. Dec",
    "mm": "2-digit month, e.g. 12",
    "dd": "2-digit day of the month, e.g. 22",
    "dow": "Day of the week as locale's full name, e.g. Tuesday",
    "doy": "Julian day of year starting from 001",
    "hour": "2-digit hour, e.g. 10",
    "min": "2-digit minute, e.g. 15",
    "sec": "2-digit second, e.g. 30",
    "strftime": "Apply strftime template to date/time. Should be used in form "
    + "{docx:created.strftime,TEMPLATE} where TEMPLATE is a valid strftime template, e.g. "
    + "{docx:created.strftime,%Y-%U} would result in year-week number of year: '2020-23'. "
    + "If used with no template will return null value. "
    + "See https://strftime.org/ for help on strftime templates.",
}


@autofile.hookimpl
def get_template_help() -> Iterable:
    text = """
    Access metadata properties of Microsoft Word document files (.docx). Use in format {docx:SUBFIELD}
    where SUBFIELD is one of the following:
 
    """
    text2 = (
        f"If the subfield is a date/time value ({', '.join(DATETIME_SUBFIELDS)}) "
        "the following attributes are available in dot notation (e.g. {docx:created.year}):"
    )

    fields = [["Field", "Description"], *[[k, v] for k, v in FIELDS.items()]]
    subfields = [["Subfield", "Description"], *[[k, v] for k, v in SUBFIELDS.items()]]
    datetime_properties = [
        ["Attribute", "Description"],
        *[[k, v] for k, v in DATETIME_ATTRIBUTES.items()],
    ]
    return [
        "**Microsoft Word Document Fields**",
        fields,
        text,
        subfields,
        text2,
        datetime_properties,
    ]


@autofile.hookimpl
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """lookup value for template docx template fields"""
    if field != "docx":
        return None

    if pathlib.Path(filepath).suffix.lower() != ".docx":
        return [None]

    subfield_parts = subfield.split(".", 1)
    if subfield_parts[0] not in SUBFIELDS:
        raise ValueError(f"Unknown docx subfield {subfield}")

    value = get_docx_property(filepath, subfield_parts[0])

    if len(subfield_parts) == 2 and isinstance(value, datetime.datetime):
        # have a date/time attribute
        dt_attribute = subfield_parts[1]
        if dt_attribute not in DATETIME_ATTRIBUTES:
            raise ValueError(f"Unknown docx datetime attribute {dt_attribute}")
        if dt_attribute == "strftime":
            if default:
                try:
                    value = value.strftime(default[0]) if value else None
                except:
                    raise ValueError(f"Invalid strftime template: '{default}'")
            else:
                value = None
            return [value]
        return [getattr(DateTimeFormatter(value), dt_attribute)]

    return [format_value(value)]


def format_value(value: Any) -> Optional[str]:
    """format values for docx template"""
    if value is None or value == "":
        return None
    elif isinstance(value, datetime.datetime):
        return value.isoformat()
    else:
        return str(value)


def get_docx_property(filepath: str, attribute: str) -> Optional[Union[List, str]]:
    """Return docx core properties attribute or None"""
    doc = docx.Document(filepath)
    core_properties = doc.core_properties
    return getattr(core_properties, attribute, None)



"""Automatically file files based on metadata templates"""
import fnmatch
import pathlib
import re
from typing import Callable, List, Optional

from .constants import NONE_STR_SENTINEL
from .pathlibutil import PathlibUtil
from .renderoptions import RenderOptions
from .template import FileTemplate, UnknownFieldError
from .utils import noop


class MultipleFilesError(Exception):
    """Multiple files found"""

    pass


def filter_file(
    filepath: pathlib.Path,
    glob: Optional[List[str]] = None,
    regex: Optional[List[str]] = None,
    filter_template: Optional[List[str]] = None,
) -> bool:
    """Return True if filepath matches one or more filters, otherwise False"""
    if not glob and not regex and not filter_template:
        return True

    glob_match = False
    if glob:
        glob_match = any(fnmatch.fnmatch(filepath.name, pattern) for pattern in glob)

    regex_match = False
    if regex:
        regex_match = any(re.search(pattern, filepath.name) for pattern in regex)

    filter_match = False
    if filter_template:
        options = RenderOptions(none_str=NONE_STR_SENTINEL)
        for pattern in filter_template:
            results, _ = FileTemplate(filepath).render(pattern, options=options)
            if results and all(NONE_STR_SENTINEL not in result for result in results):
                filter_match = True
                break

    return (
        (glob_match if glob else True)
        and (regex_match if regex else True)
        and (filter_match if filter_template else True)
    )


def process_files(
    files,
    target: str,
    directory_template: Optional[str] = None,
    filename_template: Optional[str] = None,
    walk: bool = False,
    copy: bool = False,
    hardlink: bool = False,
    dry_run: bool = False,
    glob: Optional[List[str]] = None,
    regex: Optional[List[str]] = None,
    filter_template: Optional[List[str]] = None,
    verbose: Optional[Callable] = None,
) -> int:
    """Process files"""

    verbose = verbose or noop
    dir_options = RenderOptions(dirname=True)
    file_options = RenderOptions(filename=True)

    files_processed = 0
    for filename in files:
        file = PathlibUtil(filename)
        if file.is_dir():
            if walk:
                verbose(f"Processing directory {file}")
                files_processed += process_files(
                    file.iterdir(),
                    target=target,
                    directory_template=directory_template,
                    filename_template=filename_template,
                    walk=walk,
                    copy=copy,
                    hardlink=hardlink,
                    dry_run=dry_run,
                    glob=glob,
                    regex=regex,
                    filter_template=filter_template,
                    verbose=verbose,
                )
            else:
                verbose(f"Skipping directory {file}")
        else:
            if not filter_file(file, glob, regex, filter_template):
                verbose(f"Skipping file {file}")
                continue
            target_path = pathlib.Path(target)

            rendered_directories = []
            if directory_template:
                rendered_directories, _ = FileTemplate(filename).render(
                    directory_template, options=dir_options
                )
                if len(rendered_directories) > 1 and not (any([copy, hardlink])):
                    raise MultipleFilesError(f"{rendered_directories}")

            rendered_filenames = []
            if filename_template:
                rendered_filenames, _ = FileTemplate(filename).render(
                    filename_template, options=file_options
                )
                if (len(rendered_filenames) > 1) and not (any([copy, hardlink])):
                    raise MultipleFilesError(f"{rendered_filenames}")

            # build target paths
            target_paths = [target_path]
            if rendered_directories:
                target_paths = [
                    target_path / directory for directory in rendered_directories
                ]
            if rendered_filenames:
                target_paths = [
                    target_path / filename
                    for target_path in target_paths
                    for filename in rendered_filenames
                ]
            else:
                target_paths = [target_path / file.name for target_path in target_paths]

            # move, copy, or hardlink files
            for target_path in target_paths:
                action = "Copying" if copy else "Hardlinking" if hardlink else "Moving"
                verbose(f"{action} {file} to {target_path}")
                if not dry_run:
                    if not target_path.parent.exists():
                        target_path.parent.mkdir(parents=True)
                    if hardlink:
                        file.hardlink_at(target_path)
                    elif copy:
                        file.copy_to(target_path)
                    else:
                        file.move_to(target_path)

            files_processed += 1
    return files_processed



""" Yet another simple exiftool wrapper 
    I rolled my own for following reasons: 
    1. I wanted something under MIT license (best alternative was licensed under GPL/BSD)
    2. I wanted singleton behavior so only a single exiftool process was ever running
    3. When used as a context manager, I wanted the operations to batch until exiting the context (improved performance)
    If these aren't important to you, I highly recommend you use Sven Marnach's excellent 
    pyexiftool: https://github.com/smarnach/pyexiftool which provides more functionality """

import atexit
import html
import json
import logging
import os
import re
import shutil
import subprocess
from abc import ABC, abstractmethod
from functools import lru_cache  # pylint: disable=syntax-error

# exiftool -stay_open commands outputs this EOF marker after command is run
EXIFTOOL_STAYOPEN_EOF = "{ready}"
EXIFTOOL_STAYOPEN_EOF_LEN = len(EXIFTOOL_STAYOPEN_EOF)

# list of exiftool processes to cleanup when exiting or when terminate is called
EXIFTOOL_PROCESSES = []


def escape_str(s):
    """escape string for use with exiftool -E"""
    if type(s) != str:
        return s
    s = html.escape(s)
    s = s.replace("\n", "&#xa;")
    s = s.replace("\t", "&#x9;")
    s = s.replace("\r", "&#xd;")
    return s


def unescape_str(s):
    """unescape an HTML string returned by exiftool -E"""
    if type(s) != str:
        return s
    return html.unescape(s)


@atexit.register
def terminate_exiftool():
    """Terminate any running ExifTool subprocesses; call this to cleanup when done using ExifTool"""
    for proc in EXIFTOOL_PROCESSES:
        proc._stop_proc()


@lru_cache(maxsize=1)
def get_exiftool_path():
    """return path of exiftool, cache result"""
    exiftool_path = shutil.which("exiftool")
    if exiftool_path:
        return exiftool_path.rstrip()
    else:
        raise FileNotFoundError(
            "Could not find exiftool. Please download and install from "
            "https://exiftool.org/"
        )


class _ExifToolProc:
    """Runs exiftool in a subprocess via Popen
    Creates a singleton object"""

    def __new__(cls, *args, **kwargs):
        """create new object or return instance of already created singleton"""
        if not hasattr(cls, "instance") or not cls.instance:
            cls.instance = super().__new__(cls)

        return cls.instance

    def __init__(self, exiftool=None):
        """construct _ExifToolProc singleton object or return instance of already created object
        exiftool: optional path to exiftool binary (if not provided, will search path to find it)"""

        if hasattr(self, "_process_running") and self._process_running:
            # already running
            if exiftool is not None and exiftool != self._exiftool:
                logging.warning(
                    f"exiftool subprocess already running, "
                    f"ignoring exiftool={exiftool}"
                )
            return

        self._process_running = False
        self._exiftool = exiftool or get_exiftool_path()
        self._start_proc()

    @property
    def process(self):
        """return the exiftool subprocess"""
        if self._process_running:
            return self._process
        else:
            self._start_proc()
            return self._process

    @property
    def pid(self):
        """return process id (PID) of the exiftool process"""
        return self._process.pid

    @property
    def exiftool(self):
        """return path to exiftool process"""
        return self._exiftool

    def _start_proc(self):
        """start exiftool in batch mode"""

        if self._process_running:
            logging.warning("exiftool already running: {self._process}")
            return

        # open exiftool process
        self._process = subprocess.Popen(
            [
                self._exiftool,
                "-stay_open",  # keep process open in batch mode
                "True",  # -stay_open=True, keep process open in batch mode
                "-@",  # read command-line arguments from file
                "-",  # read from stdin
                "-common_args",  # specifies args common to all commands subsequently run
                "-n",  # no print conversion (e.g. print tag values in machine readable format)
                "-P",  # Preserve file modification date/time
                "-G",  # print group name for each tag
                "-E",  # escape tag values for HTML (allows use of HTML &#xa; for newlines)
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        self._process_running = True

        EXIFTOOL_PROCESSES.append(self)

    def _stop_proc(self):
        """stop the exiftool process if it's running, otherwise, do nothing"""

        if not self._process_running:
            return

        try:
            self._process.stdin.write(b"-stay_open\n")
            self._process.stdin.write(b"False\n")
            self._process.stdin.flush()
        except Exception as e:
            pass

        try:
            self._process.communicate(timeout=5)
        except subprocess.TimeoutExpired:
            self._process.kill()
            self._process.communicate()

        del self._process
        self._process_running = False


class ExifTool:
    """Basic exiftool interface for reading and writing EXIF tags"""

    def __init__(self, filepath, exiftool=None, overwrite=True, flags=None):
        """Create ExifTool object

        Args:
            file: path to image file
            exiftool: path to exiftool, if not specified will look in path
            overwrite: if True, will overwrite image file without creating backup, default=False
            flags: optional list of exiftool flags to prepend to exiftool command when writing metadata (e.g. -m or -F)

        Returns:
            ExifTool instance
        """
        self.file = filepath
        self.overwrite = overwrite
        self.flags = flags or []
        self.data = {}
        self.warning = None
        self.error = None
        # if running as a context manager, self._context_mgr will be True
        self._context_mgr = False
        self._exiftoolproc = _ExifToolProc(exiftool=exiftool)
        self._read_exif()

    @property
    def _process(self):
        return self._exiftoolproc.process

    def setvalue(self, tag, value):
        """Set tag to value(s); if value is None, will delete tag

        Args:
            tag: str; name of tag to set
            value: str; value to set tag to

        Returns:
            True if success otherwise False

            If error generated by exiftool, returns False and sets self.error to error string
            If warning generated by exiftool, returns True (unless there was also an error) and sets self.warning to warning string
            If called in context manager, returns True (execution is delayed until exiting context manager)
        """

        if value is None:
            value = ""
        value = escape_str(value)
        command = [f"-{tag}={value}"]
        if self.overwrite and not self._context_mgr:
            command.append("-overwrite_original")

        # avoid "Warning: Some character(s) could not be encoded in Latin" warning
        command.append("-iptc:codedcharacterset=utf8")

        if self._context_mgr:
            self._commands.extend(command)
            return True
        else:
            _, _, error = self.run_commands(*command)
            return error == ""

    def addvalues(self, tag, *values):
        """Add one or more value(s) to tag
            If more than one value is passed, each value will be added to the tag

        Args:
            tag: str; tag to set
            *values: str; one or more values to set

        Returns:
            True if success otherwise False

            If error generated by exiftool, returns False and sets self.error to error string
            If warning generated by exiftool, returns True (unless there was also an error) and sets self.warning to warning string
            If called in context manager, returns True (execution is delayed until exiting context manager)

        Notes: exiftool may add duplicate values for some tags so the caller must ensure
               the values being added are not already in the EXIF data
               For some tags, such as IPTC:Keywords, this will add a new value to the list of keywords,
               but for others, such as EXIF:ISO, this will literally add a value to the existing value.
               It's up to the caller to know what exiftool will do for each tag
               If setvalue called before addvalues, exiftool does not appear to add duplicates,
               but if addvalues called without first calling setvalue, exiftool will add duplicate values
        """
        if not values:
            raise ValueError("Must pass at least one value")

        command = []
        for value in values:
            if value is None:
                raise ValueError("Can't add None value to tag")
            value = escape_str(value)
            command.append(f"-{tag}+={value}")

        if self.overwrite and not self._context_mgr:
            command.append("-overwrite_original")

        if self._context_mgr:
            self._commands.extend(command)
            return True
        else:
            _, _, error = self.run_commands(*command)
            return error == ""

    def run_commands(self, *commands, no_file=False):
        """Run commands in the exiftool process and return result.

        Args:
                *commands: exiftool commands to run
                no_file: (bool) do not pass the filename to exiftool (default=False)
                        by default, all commands will be run against self.file
                        use no_file=True to run a command without passing the filename
        Returns:
            (output, warning, errror)
            output: bytes is containing output of exiftool commands
            warning: if exiftool generated warnings, string containing warning otherwise empty string
            error: if exiftool generated errors, string containing otherwise empty string

        Note: Also sets self.warning and self.error if warning or error generated.
        """
        if not (hasattr(self, "_process") and self._process):
            raise ValueError("exiftool process is not running")

        if not commands:
            raise TypeError("must provide one or more command to run")

        if self._context_mgr and self.overwrite:
            commands = list(commands)
            commands.append("-overwrite_original")

        filename = os.fsencode(self.file) if not no_file else b""

        if self.flags:
            # need to split flags, e.g. so "--ext AVI" becomes ["--ext", "AVI"]
            flags = []
            for f in self.flags:
                flags.extend(f.split())
            command_str = b"\n".join([f.encode("utf-8") for f in flags])
            command_str += b"\n"
        else:
            command_str = b""

        command_str += (
            b"\n".join([c.encode("utf-8") for c in commands])
            + b"\n"
            + filename
            + b"\n"
            + b"-execute\n"
        )

        # send the command
        self._process.stdin.write(command_str)
        self._process.stdin.flush()

        # read the output
        output = b""
        warning = b""
        error = b""
        while EXIFTOOL_STAYOPEN_EOF not in str(output):
            line = self._process.stdout.readline()
            if line.startswith(b"Warning"):
                warning += line.strip()
            elif line.startswith(b"Error"):
                error += line.strip()
            else:
                output += line.strip()
        warning = "" if warning == b"" else warning.decode("utf-8")
        error = "" if error == b"" else error.decode("utf-8")
        self.warning = warning
        self.error = error
        return output[:-EXIFTOOL_STAYOPEN_EOF_LEN], warning, error

    @property
    def pid(self):
        """return process id (PID) of the exiftool process"""
        return self._process.pid

    @property
    def version(self):
        """returns exiftool version"""
        ver, _, _ = self.run_commands("-ver", no_file=True)
        return ver.decode("utf-8")

    def asdict(self, tag_groups=True, normalized=False):
        """return dictionary of all EXIF tags and values from exiftool
        returns empty dict if no tags

        Args:
            tag_groups: if True (default), dict keys have tag groups, e.g. "IPTC:Keywords"; if False, drops groups from keys, e.g. "Keywords"
            normalized: if True, dict keys are all normalized to lower case (default is False)
        """
        json_str, _, _ = self.run_commands("-json")
        if not json_str:
            return dict()
        json_str = unescape_str(json_str.decode("utf-8"))

        try:
            exifdict = json.loads(json_str)
        except Exception as e:
            # will fail with some commands, e.g --ext AVI which produces
            # 'No file with specified extension' instead of json
            return dict()
        exifdict = exifdict[0]
        if not tag_groups:
            # strip tag groups
            exif_new = {}
            for k, v in exifdict.items():
                k = re.sub(r".*:", "", k)
                exif_new[k] = v
            exifdict = exif_new

        if normalized:
            exifdict = {k.lower(): v for (k, v) in exifdict.items()}

        return exifdict

    def json(self):
        """returns JSON string containing all EXIF tags and values from exiftool"""
        json, _, _ = self.run_commands("-json")
        json = unescape_str(json.decode("utf-8"))
        return json

    def _read_exif(self):
        """read exif data from file"""
        data = self.asdict()
        self.data = {k: v for k, v in data.items()}

    def __str__(self):
        return f"file: {self.file}\nexiftool: {self._exiftoolproc._exiftool}"

    def __enter__(self):
        self._context_mgr = True
        self._commands = []
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            return False
        elif self._commands:
            # run_commands sets self.warning and self.error as needed
            self.run_commands(*self._commands)


class ExifToolCaching(ExifTool):
    """Basic exiftool interface for reading and writing EXIF tags, with caching.
    Use this only when you know the file's EXIF data will not be changed by any external process.

    Creates a singleton cached ExifTool instance"""

    _singletons = {}

    def __new__(cls, filepath, exiftool=None):
        """create new object or return instance of already created singleton"""
        if filepath not in cls._singletons:
            cls._singletons[filepath] = _ExifToolCaching(filepath, exiftool=exiftool)
        return cls._singletons[filepath]


class _ExifToolCaching(ExifTool):
    def __init__(self, filepath, exiftool=None):
        """Create read-only ExifTool object that caches values

        Args:
            file: path to image file
            exiftool: path to exiftool, if not specified will look in path

        Returns:
            ExifTool instance
        """
        self._json_cache = None
        self._asdict_cache = {}
        super().__init__(filepath, exiftool=exiftool, overwrite=False, flags=None)

    def run_commands(self, *commands, no_file=False):
        if commands[0] not in ["-json", "-ver"]:
            raise NotImplementedError(f"{self.__class__} is read-only")
        return super().run_commands(*commands, no_file=no_file)

    def setvalue(self, tag, value):
        raise NotImplementedError(f"{self.__class__} is read-only")

    def addvalues(self, tag, *values):
        raise NotImplementedError(f"{self.__class__} is read-only")

    def json(self):
        if not self._json_cache:
            self._json_cache = super().json()
        return self._json_cache

    def asdict(self, tag_groups=True, normalized=False):
        """return dictionary of all EXIF tags and values from exiftool
        returns empty dict if no tags

        Args:
            tag_groups: if True (default), dict keys have tag groups, e.g. "IPTC:Keywords"; if False, drops groups from keys, e.g. "Keywords"
            normalized: if True, dict keys are all normalized to lower case (default is False)
        """
        try:
            return self._asdict_cache[tag_groups][normalized]
        except KeyError:
            if tag_groups not in self._asdict_cache:
                self._asdict_cache[tag_groups] = {}
            self._asdict_cache[tag_groups][normalized] = super().asdict(
                tag_groups=tag_groups, normalized=normalized
            )
            return self._asdict_cache[tag_groups][normalized]

    def flush_cache(self):
        """Clear cached data so that calls to json or asdict return fresh data"""
        self._json_cache = None
        self._asdict_cache = {}



""" Simple formatting of datetime.datetime objects """

import datetime


class DateTimeFormatter:
    """ provides property access to formatted datetime.datetime strftime values """

    def __init__(self, dt: datetime.datetime):
        self.dt = dt

    @property
    def date(self):
        """ ISO date in form 2020-03-22 """
        return self.dt.date().isoformat()

    @property
    def year(self):
        """ 4 digit year """
        return f"{self.dt.year}"

    @property
    def yy(self):
        """ 2 digit year """
        return f"{self.dt.strftime('%y')}"

    @property
    def mm(self):
        """ 2 digit month """
        return f"{self.dt.strftime('%m')}"

    @property
    def month(self):
        """ Month as locale's full name """
        return f"{self.dt.strftime('%B')}"

    @property
    def mon(self):
        """ Month as locale's abbreviated name """
        return f"{self.dt.strftime('%b')}"

    @property
    def dd(self):
        """ 2-digit day of the month """
        return f"{self.dt.strftime('%d')}"

    @property
    def dow(self):
        """ Day of week as locale's name """
        return f"{self.dt.strftime('%A')}"

    @property
    def doy(self):
        """ Julian day of year starting from 001 """
        return f"{self.dt.strftime('%j')}"

    @property
    def hour(self):
        """ 2-digit hour """
        return f"{self.dt.strftime('%H')}"

    @property
    def min(self):
        """ 2-digit minute """
        return f"{self.dt.strftime('%M')}"

    @property
    def sec(self):
        """ 2-digit second """
        return f"{self.dt.strftime('%S')}"



""" Use template to automatically move files into directories """

import io
import pathlib
import re
import sys
from functools import partial
from typing import List, Optional

import click
from cloup import (
    Command,
    Context,
    HelpFormatter,
    HelpTheme,
    Style,
    argument,
    command,
    constraint,
    option,
    option_group,
    version_option,
)
from cloup.constraints import (
    ErrorFmt,
    If,
    RequireAtLeast,
    RequireExactly,
    mutually_exclusive,
    require_all,
)
from rich.console import Console
from rich.highlighter import NullHighlighter
from rich.markdown import Markdown
from yaspin import yaspin

from ._version import __version__
from .autofile import MultipleFilesError, process_files
from .constants import APP_NAME
from .renderoptions import RenderOptions
from .template import get_template_help
from .utils import bold, green, pluralize, red

# Set up rich console
_console = Console()
_console_stderr = Console(stderr=True)

# if True, shows verbose output, turned off via --quiet flag
_verbose = True


def verbose(message_str, **kwargs):
    if not _verbose:
        return
    _console.print(message_str, **kwargs)


def print_help_msg(command):
    with Context(command) as ctx:
        click.echo(command.get_help(ctx))


def print_error(message):
    """Print error message to stderr with rich"""
    _console_stderr.print(message, style="bold red")


def print_warning(message):
    """Print warning message to stdout with rich"""
    _console.print(message, style="bold yellow")


def echo(message):
    """print to stdout using rich"""
    _console.print(message)


# requires_one = RequireExactly(1).rephrased(
#     help="requires one",
#     error=f"it must be used with:\n" f"{ErrorFmt.param_list}",
# )


class AutofileCommand(Command):
    """Custom cloup.command that overrides get_help() to show additional help info for autofile"""

    def get_help(self, ctx):
        help_text = super().get_help(ctx)
        formatter = HelpFormatter()

        formatter.write("\n\n")
        formatter.write(rich_text(bold("Template System"), width=formatter.width))
        formatter.write("\n\n")
        for help_item in get_template_help():
            # items from get_template_help are either markdown strings or lists of lists
            if type(help_item) is str:
                formatter.write(format_markdown_str(help_item, width=formatter.width))
                formatter.write("\n")
            elif isinstance(help_item, (tuple, list)):
                help_list = [tuple(rich_text(bold(col)) for col in help_item[0])]
                help_list.extend(tuple(h) for h in help_item[1:])
                formatter.write_dl(help_list)
                formatter.write("\n")
        formatter.write_text("")
        help_text += formatter.getvalue()
        return help_text

    def get_help_option(self, ctx: Context) -> Optional["click.Option"]:
        """Returns the help option object."""
        # copied from Click source code and modified to use pager
        help_options = self.get_help_option_names(ctx)

        if not help_options or not self.add_help_option:
            return None

        def show_help(ctx: Context, param: "click.Parameter", value: str) -> None:
            if value and not ctx.resilient_parsing:
                click.echo_via_pager(ctx.get_help(), color=ctx.color)
                ctx.exit()

        return click.Option(
            help_options,
            is_flag=True,
            is_eager=True,
            expose_value=False,
            callback=show_help,
            help="Show this message and exit.",
        )


formatter_settings = HelpFormatter.settings(
    theme=HelpTheme(
        invoked_command=Style(fg="bright_yellow"),
        heading=Style(fg="bright_white", bold=True),
        constraint=Style(fg="magenta"),
        col1=Style(fg="bright_yellow"),
    )
)


@command(cls=AutofileCommand, formatter_settings=formatter_settings)
@option_group(
    "Required",
    option(
        "--target",
        "-t",
        metavar="TARGET_DIRECTORY",
        help="Target destination directory.",
        type=click.Path(exists=True, file_okay=False, resolve_path=True),
    ),
    constraint=require_all,
)
@option_group(
    "Filing templates",
    option(
        "--directory",
        "-D",
        metavar="DIRECTORY_TEMPLATE",
        help="Directory template for exporting files.",
    ),
    option(
        "--filename",
        "-F",
        metavar="FILENAME_TEMPLATE",
        help="Filename template for exporting files.",
    ),
    constraint=RequireAtLeast(1),
)
@option_group(
    "Filter Options",
    option(
        "--glob",
        "-g",
        multiple=True,
        metavar="PATTERN",
        help="Filter files to process with a glob pattern, e.g. '--glob \"*.jpg\"' "
        "--glob may be repeated to use more than one pattern. "
        'Multiple patterns treated as "OR", that is, a file that matches one or more patterns will be processed. ',
    ),
    option(
        "--regex",
        "-r",
        multiple=True,
        metavar="PATTERN",
        help="Filter files to process with a regex pattern, e.g. '--regex \"IMG_[1-3].*\"' "
        "--regex may be repeated to use more than one pattern. "
        'Multiple patterns treated as "OR", that is, a file that matches one or more patterns will be processed. '
        "Any valid python regular express may be used.",
    ),
    option(
        "--filter",
        "-f",
        "filter_template",
        multiple=True,
        metavar="TEMPLATE_PATTERN",
        help="Filter files to process that match a metadata template pattern, e.g. '--filter \"{mdls:kMDItemKind contains image}\"'. "
        "--filter matches the file if TEMPLATE_PATTERN evaluates to a non-null value. "
        "--filter may be repeated to use more than one pattern. "
        'Multiple patterns treated as "OR", that is, a file that matches one or more patterns will be processed. ',
    ),
)
@option_group(
    "Options",
    option("--walk", "-w", is_flag=True, help="Recursively walk directories."),
    option("--quiet", "-Q", is_flag=True, help="Turn off verbose output."),
    option(
        "--plain",
        is_flag=True,
        help="Plain text mode.  Do not use rich output.",
        hidden=True,
    ),
    option("--copy", "-c", is_flag=True, help="Copy files instead of moving them."),
    option(
        "--hardlink", "-h", is_flag=True, help="Hardlink files instead of moving them."
    ),
    option(
        "--dry-run",
        "-d",
        is_flag=True,
        help="Dry run mode; do not actually move/copy any files.",
    ),
    # option(
    #     "--exiftool-path",
    #     type=click.Path(exists=True),
    #     default=get_exiftool_path(),
    #     help="Optional path to exiftool executable (will look in $PATH if not specified).",
    # )
)
@constraint(mutually_exclusive, ["copy", "hardlink"])
@version_option(version=__version__)
@argument(
    "files", nargs=-1, required=True, type=click.Path(exists=True, resolve_path=True)
)
def cli(
    directory,
    filename,
    target,
    walk,
    quiet,
    plain,
    copy,
    hardlink,
    dry_run,
    glob,
    regex,
    filter_template,
    files,
):
    """move or copy files into directories based on a metadata template string"""

    # used to control whether to print out verbose output
    global _verbose
    _verbose = not quiet 

    if plain:
        # Plain text mode, disable rich output (used for testing)
        global _console
        global _console_stderr
        _console = Console(highlighter=NullHighlighter())
        _console_stderr = Console(stderr=True, highlighter=NullHighlighter())

    # create nice looking text for status
    filenames = [file for file in files if pathlib.Path(file).is_file()]
    dirnames = [file for file in files if pathlib.Path(file).is_dir()]
    text = f'Processing {len(filenames)} {pluralize(len(filenames),"file","files")}'
    dirtext = (
        f' and {len(dirnames)} {pluralize(len(dirnames), "directory","directories")}'
    )
    text = text + dirtext if walk else text
    if dirnames and not walk and not filenames:
        echo(f"Found 0 files{dirtext} but --walk was not specified, nothing to do")
        print_help_msg(cli)
        sys.exit(1)

    process_files_ = partial(
        process_files,
        target=target,
        directory_template=directory,
        filename_template=filename,
        walk=walk,
        copy=copy,
        hardlink=hardlink,
        dry_run=dry_run,
        glob=glob,
        regex=regex,
        filter_template=filter_template,
        verbose=verbose,
    )

    if not _verbose:
        with yaspin(text=text):
            try:
                files_processed = process_files_(files)
            except MultipleFilesError as e:
                print_error(
                    "Error: --directory or --filename template produced multiple target paths; cannot move file to more than one target path. "
                    "Change the template or use --copy or --hardlink: "
                    f"{e}"
                )
                sys.exit(1)
    else:
        verbose(text)
        try:
            files_processed = process_files_(files)
        except MultipleFilesError as e:
            print_error(
                "Error: --directory or --filename template produced multiple target paths; cannot move file to more than one target path. "
                "Change the template or use --copy or --hardlink: "
                f"{e}"
            )
            sys.exit(1)

    verbose(
        f"Done. Processed {files_processed} {pluralize(files_processed, 'file', 'files')}."
    )

    echo("Done.")


def rich_text(text, width=78):
    """Return rich formatted text"""
    sio = io.StringIO()
    console = Console(file=sio, force_terminal=True, width=width)
    console.print(text)
    rich_text = sio.getvalue()
    rich_text = rich_text.rstrip()
    sio.close()
    return rich_text


def strip_md_header_and_links(md):
    """strip markdown headers and links from markdown text md

    Args:
        md: str, markdown text

    Returns:
        str with markdown headers and links removed

    Note: This uses a very basic regex that likely fails on all sorts of edge cases
    but works for the links in the docs
    """
    links = r"(?:[*#])|\[(.*?)\]\(.+?\)"

    def subfn(match):
        return match.group(1)

    return re.sub(links, subfn, md)


def strip_md_links(md):
    """strip markdown links from markdown text md

    Args:
        md: str, markdown text

    Returns:
        str with markdown links removed

    Note: This uses a very basic regex that likely fails on all sorts of edge cases
    but works for the links in the osxphotos docs
    """
    links = r"\[(.*?)\]\(.+?\)"

    def subfn(match):
        return match.group(1)

    return re.sub(links, subfn, md)


def strip_html_comments(text):
    """Strip html comments from text (which doesn't need to be valid HTML)"""
    return re.sub(r"<!--(.|\s|\n)*?-->", "", text)


def format_markdown_str(string, width=78):
    """Return formatted markdown str for terminal"""
    sio = io.StringIO()
    console = Console(file=sio, force_terminal=True, width=width)
    console.print(Markdown(string))
    help_str = sio.getvalue()
    sio.close()
    return help_str


def main():
    cli()



from typing import List, Optional, Iterable

from pluggy import HookspecMarker

hookspec = HookspecMarker("autofile")


@hookspec(firstresult=True)
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """Called by template.py to get template value for custom template

    Return: None if field is not handled by this plugin otherwise list of str values"""

    # return value of None means that field is not handled by this plugin
    # return value of [None] means that field is handled by this plugin but value resolved to None (no value)
    # return value of [value] means that field is handled by this plugin and value resolved to value


@hookspec
def get_template_help() -> Iterable:
    """Return iterable of one or more help elements. Each element may be a str, a dict, or a list of lists"""



"""File stat template plugin for autofile"""

import grp
import os
import pwd
from typing import Iterable, List, Optional

import autofile

FIELDS = {
    "{size}": "Size of file in bytes",
    "{uid}": "User identifier of the file owner",
    "{gid}": "Group identifier of the file owner",
    "{user}": "User name of the file owner",
    "{group}": "Group name of the file owner",
}


@autofile.hookimpl
def get_template_help() -> Iterable:
    fields = [["Field", "Description"], *[[k, v] for k, v in FIELDS.items()]]
    return ["**File Information Fields**", fields]


@autofile.hookimpl
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """lookup value for os.stat values for filepath

    Args:
        field: template field to find value for.

    Returns:
        The matching template value (which may be None).
    """
    if "{" + field + "}" not in FIELDS:
        return None

    stat_info = os.stat(filepath)
    val = None
    if field == "size":
        val = stat_info.st_size
    elif field == "uid":
        val = stat_info.st_uid
    elif field == "gid":
        val = stat_info.st_gid
    elif field == "user":
        val = pwd.getpwuid(stat_info.st_uid).pw_name
    elif field == "group":
        val = grp.getgrgid(stat_info.st_gid).gr_name
    else:
        raise ValueError("Unknown field: {" + field + "}")

    return [str(val)]



"""RenderOptions class for template system"""

from dataclasses import dataclass
from typing import Optional

from .constants import INPLACE_DEFAULT, NONE_STR_DEFAULT, NONE_STR_SENTINEL


@dataclass
class RenderOptions:
    """Options for PhotoTemplate.render

    tag: tag name being processed
    none_str: str to use default for None values
    path_sep: optional string to use as path separator, default is os.path.sep
    expand_inplace: expand multi-valued substitutions in-place as a single string
        instead of returning individual strings
    inplace_sep: optional string to use as separator between multi-valued keywords
    with expand_inplace; default is ','
    filename: if True, template output will be sanitized to produce valid file name
    dirname: if True, template output will be sanitized to produce valid directory name
    strip: if True, strips leading/trailing whitespace from rendered templates
    dest_path: set to the destination path of the file (for use by {function} template)
    quote: quote path templates for execution in the shell
    """

    tag: Optional[str] = None
    none_str: str = NONE_STR_DEFAULT
    expand_inplace: bool = False
    inplace_sep: Optional[str] = INPLACE_DEFAULT
    filename: bool = False
    dirname: bool = False
    strip: bool = False
    dest_path: Optional[str] = None
    quote: bool = False



"""autofile plugin to read PDF metadata"""

import datetime
import logging
import pathlib
import re
from typing import Any, Iterable, List, Optional, Union

from pdfminer.pdfdocument import PDFDocument
from pdfminer.pdfparser import PDFParser

import autofile
from autofile.datetime_formatter import DateTimeFormatter

# pdfminer.six logs a crazy amount of info to logging.INFO so turn off the noise
logging.getLogger("pdfminer").setLevel(logging.WARNING)

FIELDS = {
    "{pdf}": "Access metadata properties of Adobe PDF files (.pdf); "
    "use in format {pdf:SUBFIELD}"
}

SUBFIELDS = {
    "author": "Author of the document.",
    "creator": "The application that created the document.",
    "producer": "The application the produced the PDF (may be different than creator).",
    "created": "Date of creation of the document; a date/time value.",
    "modified": "Date on which the document was changed; a date/time value.",
    "subject": "The topic of the content of the document.",
    "title": "The name given to the document.",
    "keywords": "Keywords associated with the document; a string of delimited words.",
}

SUBFIELD_MAPPING = {
    "author": "Author",
    "creator": "Creator",
    "producer": "Producer",
    "created": "CreationDate",
    "modified": "ModDate",
    "subject": "Subject",
    "title": "Title",
    "keywords": "Keywords",
}

DATETIME_SUBFIELDS = ["created", "modified"]

DATETIME_ATTRIBUTES = {
    "date": "ISO date, e.g. 2020-03-22",
    "year": "4-digit year, e.g. 2021",
    "yy": "2-digit year, e.g. 21",
    "month": "Month name as locale's full name, e.g. December",
    "mon": "Month as locale's abbreviated name, e.g. Dec",
    "mm": "2-digit month, e.g. 12",
    "dd": "2-digit day of the month, e.g. 22",
    "dow": "Day of the week as locale's full name, e.g. Tuesday",
    "doy": "Julian day of year starting from 001",
    "hour": "2-digit hour, e.g. 10",
    "min": "2-digit minute, e.g. 15",
    "sec": "2-digit second, e.g. 30",
    "strftime": "Apply strftime template to date/time. Should be used in form "
    + "{docx:created.strftime,TEMPLATE} where TEMPLATE is a valid strftime template, e.g. "
    + "{docx:created.strftime,%Y-%U} would result in year-week number of year: '2020-23'. "
    + "If used with no template will return null value. "
    + "See https://strftime.org/ for help on strftime templates.",
}


@autofile.hookimpl
def get_template_help() -> Iterable:
    text = """
    Access metadata properties of Adobe PDF files (.pdf). Use in format {pdf:SUBFIELD}
    where SUBFIELD is one of the following:
 
    """
    text2 = (
        f"If the subfield is a date/time value ({', '.join(DATETIME_SUBFIELDS)}) "
        "the following attributes are available in dot notation (e.g. {pdf:created.year}):"
    )

    fields = [["Field", "Description"], *[[k, v] for k, v in FIELDS.items()]]
    subfields = [["Subfield", "Description"], *[[k, v] for k, v in SUBFIELDS.items()]]
    datetime_properties = [
        ["Attribute", "Description"],
        *[[k, v] for k, v in DATETIME_ATTRIBUTES.items()],
    ]
    return [
        "**Adobe PDF Document Fields**",
        fields,
        text,
        subfields,
        text2,
        datetime_properties,
    ]


@autofile.hookimpl
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """lookup value for template pdf template fields"""
    if field != "pdf":
        return None

    if pathlib.Path(filepath).suffix.lower() != ".pdf":
        return [None]

    subfield_parts = subfield.split(".", 1)
    if subfield_parts[0] not in SUBFIELDS:
        raise ValueError(f"Unknown pdf subfield {subfield}")

    value = get_pdf_property(filepath, subfield_parts[0])

    if len(subfield_parts) == 2 and isinstance(value, datetime.datetime):
        # have a date/time attribute
        dt_attribute = subfield_parts[1]
        if dt_attribute not in DATETIME_ATTRIBUTES:
            raise ValueError(f"Unknown docx datetime attribute {dt_attribute}")
        if dt_attribute == "strftime":
            if default:
                try:
                    value = value.strftime(default[0]) if value else None
                except:
                    raise ValueError(f"Invalid strftime template: '{default}'")
            else:
                value = None
            return [value]
        return [getattr(DateTimeFormatter(value), dt_attribute)]

    return [format_value(value)]


def format_value(value: Any) -> Optional[str]:
    """format values for pdf template"""
    if not value:
        return None
    elif isinstance(value, datetime.datetime):
        return value.isoformat()
    else:
        return str(value)


def get_pdf_property(
    filepath: str, subfield: str
) -> Optional[Union[str, datetime.datetime]]:
    """get pdf property"""
    metadata = get_pdf_metadata_info(filepath)
    try:
        subfield_mapping = SUBFIELD_MAPPING[subfield]
    except KeyError:
        raise ValueError(f"Unknown pdf subfield {subfield}")
    value = metadata.get(subfield_mapping, None)
    if not value:
        return None

    if subfield in DATETIME_SUBFIELDS:
        return pdf_date_to_datetime(decode_pdf_field(value))
    return decode_pdf_field(value)


def decode_pdf_field(field: Any) -> str:
    """decode pdf metadata field from binary unicode to str"""
    # XMP Spec says that PDF metadata fields are encoded in UTF-8, but
    # that is not always the case so this tries various encodings to find one that works
    for encoding in ("utf-8", "utf-16", "utf-32"):
        try:
            value = field.decode(encoding)
        except UnicodeDecodeError:
            pass
        else:
            return value
    return field.decode("utf-8", errors="ignore")


def get_pdf_metadata_info(pdfpath: str) -> dict:
    """
    Get the metadata info from a pdf file
    """
    with open(pdfpath, "rb") as fp:
        parser = PDFParser(fp)
        doc = PDFDocument(parser)
        return doc.info[0] if doc.info else {}


def pdf_date_to_datetime(date_str: str) -> datetime.datetime:
    """
    Convert a pdf date such as "D:20120321183444+07'00'" into a usable datetime
    http://www.verypdf.com/pdfinfoeditor/pdf-date-format.htm
    (D:YYYYMMDDHHmmSSOHH'mm')
    From https://stackoverflow.com/questions/16503075/convert-creationtime-of-pdf-to-a-readable-format-in-python
    :param date_str: pdf date string
    :return: datetime object
    """

    pdf_date_pattern = re.compile(
        "".join(
            [
                r"(D:)?",
                r"(?P<year>\d\d\d\d)",
                r"(?P<month>\d\d)",
                r"(?P<day>\d\d)",
                r"(?P<hour>\d\d)",
                r"(?P<minute>\d\d)",
                r"(?P<second>\d\d)",
                r"(?P<tz_offset>[+-zZ])?",
                r"(?P<tz_hour>\d\d)?",
                r"'?(?P<tz_minute>\d\d)?'?",
            ]
        )
    )

    match = re.match(pdf_date_pattern, date_str)
    if match:
        date_info = match.groupdict()

        for k, v in date_info.items():  # transform values
            if v is None:
                pass
            elif k == "tz_offset":
                date_info[k] = v.lower()  # so we can treat Z as z
            else:
                date_info[k] = int(v)

        if date_info["tz_offset"] in ("z", None):  # UTC
            date_info["tzinfo"] = datetime.timezone.utc
        else:
            multiplier = 1 if date_info["tz_offset"] == "+" else -1
            date_info["tzinfo"] = datetime.timezone(
                datetime.timedelta(
                    seconds=multiplier
                    * (3600 * date_info["tz_hour"] + 60 * date_info["tz_minute"]),
                )
            )

        for k in ("tz_offset", "tz_hour", "tz_minute"):  # no longer needed
            del date_info[k]

        return datetime.datetime(**date_info)



""" FileUtil class with methods for copy, hardlink, unlink, etc. Works with pathlib.Path """

import pathlib
import platform
import shutil
import os

# python version as 2-digit float (e.g. 3.7)
PY_VERSION = float(".".join(platform.python_version_tuple()[:2]))
SYSTEM = platform.system()
PYOBJC = False

if SYSTEM == "Darwin":
    try:
        import Foundation

        PYOBJC = True
    except ImportError:
        pass


class PathlibUtil:
    """Various file utilities based on pathlib.Path"""

    def __init__(self, path):
        self.path = pathlib.Path(path)

    def __getattr__(self, attr):
        # delegate to pathlib.Path
        return_value = getattr(self.path, attr)
        if isinstance(return_value, pathlib.Path):
            return PathlibUtil(return_value)
        return return_value

    def copy_to(self, dest):
        """Copies a file or directory from src path to dest path

        Args:
            dest: destination path as string
                  dest may be either directory or file; in either case, src file/directory must not exist in dest
            Note: src and dest may be either a string or a pathlib.Path object

        Returns:
            PathLibUtil object of the new file/directory

        Raises:
            OSError if copy fails
            FileExistsError if dest already exists
            FileNotFoundError if src does not exist
        """
        if not isinstance(dest, PathlibUtil):
            dest = PathlibUtil(dest)

        if not self.path.exists():
            raise FileNotFoundError(f"{self.path} does not exist")

        if dest.is_file():
            raise FileExistsError(f"{dest} already exists")

        if dest.is_dir():
            dest = dest / self.path.name

        if self._pyobjc:
            return self._copy_to_mac(dest)

        if self.path.is_dir():
            shutil.copytree(str(self.path), str(dest))
        else:
            shutil.copy2(str(self.path), str(dest))

        return dest

    def _copy_to_mac(self, dest):
        """copy_to implementation that uses NSFileManager to take advantage of copy on write for APFS

        If the source is a directory, this method copies the directory and all of its contents, including any hidden files
        """
        src = self.path
        error = Foundation.NSFileManager.defaultManager().copyItemAtPath_toPath_error_(
            str(src), str(dest), None
        )
        # error is a tuple of (bool, error_string)
        # error[0] is True if copy succeeded
        if not error[0]:
            raise OSError(error[1])
        return dest

    def move_to(self, dest):
        """Move file or directory to dest"""
        if not isinstance(dest, PathlibUtil):
            dest = PathlibUtil(dest)

        if not self.path.exists():
            raise FileNotFoundError(f"{self.path} does not exist")

        if dest.is_file():
            raise FileExistsError(f"{dest} already exists")

        if dest.is_dir():
            dest = dest / self.path.name

        if self._pyobjc:
            return self._move_to_mac(dest)

        shutil.move(str(self.path), str(dest))
        return dest

    def _move_to_mac(self, dest):
        """move_to implementation that uses NSFileManager

        If the source is a directory, this method moves the directory and all of its contents, including any hidden files
        """
        src = self.path
        error = Foundation.NSFileManager.defaultManager().moveItemAtPath_toPath_error_(
            str(src), str(dest), None
        )
        # error is a tuple of (bool, error_string)
        # error[0] is True if copy succeeded
        if not error[0]:
            raise OSError(error[1])
        return dest

    def mkdir(self):
        """Create directory at this path"""
        # pathlib.Path().mkdir returns None but we want to return the new path
        self.path.mkdir()
        return self

    def hardlink_to(self, target):
        """Make this path a hard link to the same file as target (opposite of hardlink_at)"""
        # This is included in Python 3.10 but not earlier
        if not isinstance(target, PathlibUtil):
            target = PathlibUtil(target)

        if self.path.exists():
            raise FileExistsError(f"{self.path} already exists")

        if not target.exists():
            raise FileNotFoundError(f"{target} does not exist")

        os.link(str(target), str(self.path))
        return self

    def hardlink_at(self, dest):
        """Make dest a hardlink to this path (opposite of hardlink_to)"""
        if not isinstance(dest, PathlibUtil):
            dest = PathlibUtil(dest)

        if dest.exists():
            raise FileExistsError(f"{dest} already exists")

        if not self.path.exists():
            raise FileNotFoundError(f"{self.path} does not exist")

        return dest.hardlink_to(self.path)

    @property
    def _pyobjc(self):
        return PYOBJC

    def __str__(self):
        return self.path.__str__()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.path})"

    def __truediv__(self, key):
        try:
            return self.path._make_child((key,))
        except TypeError:
            return NotImplemented

    def __rtruediv__(self, key):
        try:
            return self.path._from_parts([key] + self._parts)
        except TypeError:
            return NotImplemented

    # @classmethod
    # def unlink(cls, filepath):
    #     """unlink filepath; if it's pathlib.Path, use Path.unlink, otherwise use os.unlink"""
    #     if isinstance(filepath, pathlib.Path):
    #         filepath.unlink()
    #     else:
    #         os.unlink(filepath)

    # @classmethod
    # def rmdir(cls, dirpath):
    #     """remove directory filepath; dirpath must be empty"""
    #     if isinstance(dirpath, pathlib.Path):
    #         dirpath.rmdir()
    #     else:
    #         os.rmdir(dirpath)

    # @classmethod
    # def utime(cls, path, times):
    #     """Set the access and modified time of path."""
    #     os.utime(path, times)

    # @classmethod
    # def cmp(cls, f1, f2, mtime1=None):
    #     """Does shallow compare (file signatures) of f1 to file f2.
    #     Arguments:
    #     f1 --  File name
    #     f2 -- File name
    #     mtime1 -- optional, pass alternate file modification timestamp for f1; will be converted to int

    #     Return value:
    #     True if the file signatures as returned by stat are the same, False otherwise.
    #     Does not do a byte-by-byte comparison.
    #     """

    #     s1 = cls._sig(os.stat(f1))
    #     if mtime1 is not None:
    #         s1 = (s1[0], s1[1], int(mtime1))
    #     s2 = cls._sig(os.stat(f2))
    #     if s1[0] != stat.S_IFREG or s2[0] != stat.S_IFREG:
    #         return False
    #     return s1 == s2

    # @classmethod
    # def cmp_file_sig(cls, f1, s2):
    #     """Compare file f1 to signature s2.
    #     Arguments:
    #     f1 --  File name
    #     s2 -- stats as returned by _sig

    #     Return value:
    #     True if the files are the same, False otherwise.
    #     """

    #     if not s2:
    #         return False

    #     s1 = cls._sig(os.stat(f1))

    #     if s1[0] != stat.S_IFREG or s2[0] != stat.S_IFREG:
    #         return False
    #     return s1 == s2

    # @classmethod
    # def file_sig(cls, f1):
    #     """return os.stat signature for file f1"""
    #     return cls._sig(os.stat(f1))

    # @classmethod
    # def rename(cls, src, dest):
    #     """Copy src to dest

    #     Args:
    #         src: path to source file
    #         dest: path to destination file

    #     Returns:
    #         Name of renamed file (dest)

    #     """
    #     os.rename(str(src), str(dest))
    #     return dest

    # @staticmethod
    # def _sig(st):
    #     """return tuple of (mode, size, mtime) of file based on os.stat
    #     Args:
    #         st: os.stat signature
    #     """
    #     # use int(st.st_mtime) because ditto does not copy fractional portion of mtime
    #     return (stat.S_IFMT(st.st_mode), st.st_size, int(st.st_mtime))



""" Custom template system for autofile """

import importlib
import locale
import pathlib
import re
import shlex
from textwrap import dedent
from typing import Dict, List, Optional, Tuple, Union

import pluggy
from textx import TextXSyntaxError, metamodel_from_file

from . import hookspecs
from ._version import __version__
from .constants import APP_NAME
from .path_utils import sanitize_dirname, sanitize_filename, sanitize_pathpart
from .pathlibutil import PathlibUtil
from .renderoptions import RenderOptions


class UnknownFieldError(Exception):
    """Raised when a field is not known"""

    pass


class SyntaxError(Exception):
    """Raised when template engine cannot parse the template string"""

    pass


DEFAULT_PLUGINS = (
    "autofile.plugins.templates.punctuation",
    "autofile.plugins.templates.docx",
    "autofile.plugins.templates.pdf",
    "autofile.plugins.templates.exiftool",
    "autofile.plugins.templates.audio",
    "autofile.plugins.templates.uti",
    "autofile.plugins.templates.finder",
    "autofile.plugins.templates.mdls",
    "autofile.plugins.templates.format",
    "autofile.plugins.templates.filepath",
    "autofile.plugins.templates.filedates",
    "autofile.plugins.templates.filestat",
)


def get_plugin_manager():
    pm = pluggy.PluginManager(APP_NAME)
    pm.add_hookspecs(hookspecs)
    pm.load_setuptools_entrypoints(APP_NAME)
    return pm


PM = get_plugin_manager()

# Load default plugins
for plugin in DEFAULT_PLUGINS:
    mod = importlib.import_module(plugin)
    PM.register(mod, plugin)

# ensure locale set to user's locale
locale.setlocale(locale.LC_ALL, "")

MTL_GRAMMAR_MODEL = str(pathlib.Path(__file__).parent / "template.tx")
"""TextX metamodel for template language """

# # Permitted multi-value substitutions (each of these returns None or 1 or more values)
# TEMPLATE_SUBSTITUTIONS_MULTI_VALUED = {
#     # "{shell_quote}": "Use in form '{shell_quote,TEMPLATE}'; quotes the rendered TEMPLATE value(s) for safe usage in the shell, e.g. My file.jpeg => 'My file.jpeg'; only adds quotes if needed.",
#     # "{function}": "Execute a python function from an external file and use return value as template substitution. "
#     # + "Use in format: {function:file.py::function_name} where 'file.py' is the name of the python file and 'function_name' is the name of the function to call. "
#     # + "The function will be passed the PhotoInfo object for the photo. "
#     # + "See https://github.com/RhetTbull/osxphotos/blob/master/examples/template_function.py for an example of how to implement a template function.",
# }


FILTER_VALUES = {
    "lower": "Convert value to lower case, e.g. 'Value' => 'value'.",
    "upper": "Convert value to upper case, e.g. 'Value' => 'VALUE'.",
    "strip": "Strip whitespace from beginning/end of value, e.g. ' Value ' => 'Value'.",
    "titlecase": "Convert value to title case, e.g. 'my value' => 'My Value'.",
    "capitalize": "Capitalize first word of value and convert other words to lower case, e.g. 'MY VALUE' => 'My value'.",
    "braces": "Enclose value in curly braces, e.g. 'value => '{value}'.",
    "parens": "Enclose value in parentheses, e.g. 'value' => '(value').",
    "brackets": "Enclose value in brackets, e.g. 'value' => '[value]'.",
    "shell_quote": "Quotes the value for safe usage in the shell, e.g. My file.jpeg => 'My file.jpeg'; only adds quotes if needed.",
    "autosplit": "Automatically split delimited string into separate values (for example, keyword string in docx files); will split strings delimited by comma, semicolon, or space.",
    "split(str)": "Splits the value into a list using 'str' as delimiter, e.g. split(;) would split 'foo;bar' into [foo, bar].",
    "chop(x)": "Remove x characters off the end of value, e.g. chop(1): 'Value' => 'Valu'.",
    "chomp(x)": "Remove x characters from the beginning of value, e.g. chomp(1): 'Value' => 'alue'.",
    # "function": "Run custom python function to filter value; use in format 'function:/path/to/file.py::function_name'. See example at https://github.com/RhetTbull/osxphotos/blob/master/examples/template_filter.py",
}

# PATH_SEP_DEFAULT = os.path.sep


class FileTemplateParser:
    """Parser for FileTemplate"""

    # implemented as Singleton

    def __new__(cls, *args, **kwargs):
        """create new object or return instance of already created singleton"""
        if not hasattr(cls, "instance") or not cls.instance:
            cls.instance = super().__new__(cls)

        return cls.instance

    def __init__(self):
        """return existing singleton or create a new one"""

        if hasattr(self, "metamodel"):
            return

        self.metamodel = metamodel_from_file(MTL_GRAMMAR_MODEL, skipws=False)

    def parse(self, template_statement):
        """Parse a template_statement string"""
        return self.metamodel.model_from_str(template_statement)

    def fields(self, template_statement):
        """Return list of fields found in a template statement; does not verify that fields are valid"""
        model = self.parse(template_statement)
        return [ts.template.field for ts in model.template_strings if ts.template]


class FileTemplate:
    """FileTemplate class to render a template string from a file and it's associated metadata"""

    def __init__(self, filepath: Union[str, PathlibUtil, pathlib.Path]):
        """Inits FileTemplate class with filepath and exiftool_path"""

        if isinstance(filepath, str):
            filepath = PathlibUtil(filepath)
        if not filepath.exists():
            raise FileNotFoundError(f"File {filepath} does not exist")

        self.filepath = str(filepath)
        self.hook = PM.hook

        # get parser singleton
        self.parser = FileTemplateParser()

        # initialize render options
        # this will be done in render() but for testing, some of the lookup functions are called directly
        options = RenderOptions()
        self.options = options
        self.tag = options.tag
        self.group, self.tagname = "", ""
        self.inplace_sep = options.inplace_sep
        self.none_str = options.none_str
        self.expand_inplace = options.expand_inplace
        self.filename = options.filename
        self.dirname = options.dirname
        self.strip = options.strip
        self.quote = options.quote
        self.dest_path = options.dest_path
        # self.exiftool = options.exiftool or ExifToolCaching(
        #     self.filepath, exiftool=self.exiftool_path
        # )
        self.variables = {}

    def render(
        self,
        template: str,
        options: Optional[RenderOptions] = None,
    ):
        """Render a filename or directory template

        Args:
            template: str template
            options: a RenderOptions instance

        Returns:
            ([rendered_strings], [unmatched]): tuple of list of rendered strings and list of unmatched template values
        """

        if type(template) is not str:
            raise TypeError(f"template must be type str, not {type(template)}")

        self.variables = {}

        options = options or RenderOptions()
        self.options = options
        self.tag = options.tag
        self.inplace_sep = options.inplace_sep
        self.none_str = options.none_str
        self.expand_inplace = options.expand_inplace
        self.filename = options.filename
        self.dirname = options.dirname
        self.strip = options.strip
        self.dest_path = options.dest_path
        self.quote = options.quote
        self.dest_path = options.dest_path

        try:
            model = self.parser.parse(template)
        except TextXSyntaxError as e:
            raise SyntaxError(e)

        if not model:
            # empty string
            return [], []

        return self._render_statement(model)

    def _render_statement(
        self,
        statement,
    ):
        results = []
        unmatched = []
        for ts in statement.template_strings:
            results, unmatched = self._render_template_string(
                ts,
                results=results,
                unmatched=unmatched,
            )

        rendered_strings = results

        if self.filename:
            rendered_strings = [
                sanitize_filename(rendered_str) for rendered_str in rendered_strings
            ]

        if self.strip:
            rendered_strings = [
                rendered_str.strip() for rendered_str in rendered_strings
            ]

        return rendered_strings, unmatched

    def _render_template_string(
        self,
        ts,
        results=None,
        unmatched=None,
    ):
        """Render a TemplateString object"""

        results = results or [""]
        unmatched = unmatched or []

        if ts.template:
            # have a template field to process
            field = ts.template.field
            subfield = ts.template.subfield

            # process filters
            filters = []
            if ts.template.filter is not None:
                filters = ts.template.filter.value

            # process delim
            if ts.template.delim is not None:
                # if value is None, means format was {+field}
                delim = ts.template.delim.value or ""
            else:
                delim = None

            if ts.template.bool is not None:
                is_bool = True
                if ts.template.bool.value is not None:
                    bool_val, u = self._render_statement(
                        ts.template.bool.value,
                    )
                    unmatched.extend(u)
                else:
                    # blank bool value
                    bool_val = [""]
            else:
                is_bool = False
                bool_val = None

            # process default
            if ts.template.default is not None:
                # default is also a TemplateString
                if ts.template.default.value is not None:
                    default, u = self._render_statement(
                        ts.template.default.value,
                    )
                    unmatched.extend(u)
                else:
                    # blank default value
                    default = [""]
            else:
                default = []

            # process conditional
            if ts.template.conditional is not None:
                operator = ts.template.conditional.operator
                negation = ts.template.conditional.negation
                if ts.template.conditional.value is not None:
                    # conditional value is also a TemplateString
                    conditional_value, u = self._render_statement(
                        ts.template.conditional.value,
                        # path_sep=path_sep,
                    )
                    unmatched.extend(u)
                else:
                    # this shouldn't happen
                    conditional_value = [""]
            else:
                operator = None
                negation = None
                conditional_value = []

            # vals = []
            # # elif field == "function":
            # #     if subfield is None:
            # #         raise ValueError(
            # #             "SyntaxError: filename and function must not be null with {function::filename.py:function_name}"
            # #         )
            # #     vals = self.get_template_value_function(
            # #         subfield,
            # #     )

            if field.startswith("%"):
                # variable in form {%var}
                vals = self.variables.get(field[1:], None)
                if vals is None:
                    raise SyntaxError(f"Variable '{field[1:]}' is not defined.")
            elif field == "var":
                if not subfield or not default:
                    raise SyntaxError(
                        "var must have a subfield and value in form {var:subfield,value}"
                    )
                self.variables[subfield] = default
                vals = []
            else:
                # pass processing to plugins to get values
                vals = self.hook.get_template_value(
                    filepath=self.filepath,
                    field=field,
                    subfield=subfield,
                    default=default,
                    options=self.options,
                )

            if vals:
                if self.filename:
                    vals = [sanitize_pathpart(v) for v in vals]
                elif self.dirname:
                    vals = [sanitize_dirname(v) for v in vals]

            if vals is None:
                if field:
                    raise UnknownFieldError(f"Unknown template field: {field}")
                vals = []

            vals = [val for val in vals if val is not None]

            if self.expand_inplace or delim is not None:
                sep = delim if delim is not None else self.inplace_sep
                vals = [sep.join(sorted(vals))] if vals else []

            for filter_ in filters:
                vals = self.get_template_value_filter(filter_, vals)

            # process find/replace
            if ts.template.findreplace:
                new_vals = []
                for val in vals:
                    for pair in ts.template.findreplace.pairs:
                        find = pair.find or ""
                        find_vars = self.expand_variables(find)
                        if len(find_vars) != 1:
                            raise SyntaxError(
                                f"find/replace must have a single value to find or replace: {find_vars}"
                            )
                        find = find_vars[0]

                        repl = pair.replace or ""
                        repl_vars = self.expand_variables(repl)
                        if len(repl_vars) != 1:
                            raise SyntaxError(
                                f"find/replace must have a single value to replace: {repl_vars}"
                            )
                        repl = repl_vars[0]
                        val = val.replace(find, repl)
                    new_vals.append(val)
                vals = new_vals

            if operator:
                # have a conditional operator

                def string_test(test_function):
                    """Perform string comparison using test_function; closure to capture conditional_value, vals, negation"""
                    match = False
                    for c in conditional_value:
                        for v in vals:
                            if test_function(v, c):
                                match = True
                                break
                        if match:
                            break
                    if (match and not negation) or (negation and not match):
                        return ["True"]
                    else:
                        return []

                def comparison_test(test_function):
                    """Perform numerical comparisons using test_function; closure to capture conditional_val, vals, negation"""
                    # returns True if any of the values match the condition
                    if len(conditional_value) != 1:
                        raise SyntaxError(
                            f"comparison operators may only be used with a single conditional value: {conditional_value}"
                        )
                    try:
                        match = any(
                            bool(test_function(float(v), float(conditional_value[0])))
                            for v in vals
                        )
                        if (match and not negation) or (negation and not match):
                            return ["True"]
                        else:
                            return []
                    except ValueError as e:
                        raise SyntaxError(
                            f"comparison operators may only be used with values that can be converted to numbers: {vals} {conditional_value}"
                        )

                if operator in ["contains", "matches", "startswith", "endswith"]:
                    # process any "or" values separated by "|"
                    temp_values = []
                    for c in conditional_value:
                        temp_values.extend(c.split("|"))
                    conditional_value = temp_values

                if operator == "contains":
                    vals = string_test(lambda v, c: c in v)
                elif operator == "matches":
                    vals = string_test(lambda v, c: v == c)
                elif operator == "startswith":
                    vals = string_test(lambda v, c: v.startswith(c))
                elif operator == "endswith":
                    vals = string_test(lambda v, c: v.endswith(c))
                elif operator == "==":
                    match = sorted(vals) == sorted(conditional_value)
                    if (match and not negation) or (negation and not match):
                        vals = ["True"]
                    else:
                        vals = []
                elif operator == "!=":
                    match = sorted(vals) != sorted(conditional_value)
                    if (match and not negation) or (negation and not match):
                        vals = ["True"]
                    else:
                        vals = []
                elif operator == "<":
                    vals = comparison_test(lambda v, c: v < c)
                elif operator == "<=":
                    vals = comparison_test(lambda v, c: v <= c)
                elif operator == ">":
                    vals = comparison_test(lambda v, c: v > c)
                elif operator == ">=":
                    vals = comparison_test(lambda v, c: v >= c)

            if is_bool:
                vals = default if not vals else bool_val
            elif not vals and field != "var":
                # don't assign default value if the template was variable assignment
                vals = default or [self.none_str]

            pre = ts.pre or ""
            post = ts.post or ""

            rendered = [pre + val + post for val in vals]
            results_new = []
            for ren in rendered:
                for res in results:
                    res_new = res + ren
                    results_new.append(res_new)
            results = results_new

        else:
            # no template
            pre = ts.pre or ""
            post = ts.post or ""
            results = [r + pre + post for r in results]

        return results, unmatched

    def expand_variables(self, value: str) -> List[str]:
        """Expand variables in value"""
        # replace any variables with their values
        values = [value]
        new_values = []
        # allow %% to escape %, match variables in form %var
        variable_match = re.compile(r"(?:%%)*(%[\w]+)?")
        while True:
            for value in values:
                match = variable_match.search(value)
                if not match or not match.group(1):
                    break
                var = match.group(1)
                var_name = var[1:]
                if var_name not in self.variables:
                    raise SyntaxError(f"Variable '{var_name}' is not defined.")
                for val in values:
                    for var_val in self.variables[var_name]:
                        new_values.append(
                            re.sub(r"(%%)*" + f"{var}", r"\g<1>" + var_val, val)
                        )
            if new_values == values or not new_values:
                break
            values = new_values.copy()
            new_values = []

        # replace %% with %
        # any %% left in the string will be replaced with %
        values = [value.replace("%%", "%") for value in values]

        return values

    def get_template_value_filter(self, filter_, values):
        if re.search(r"\(.*\)", filter_):
            # filter has arguments
            filter_, args = filter_.split("(", 1)
            args = args.rstrip(")")
            args = self.expand_variables(args)
            if len(args) != 1:
                raise SyntaxError(f"Filter arguments must be a single value: {args}")
            args = args[0]
        else:
            args = None
        if filter_ == "lower":
            if values and type(values) == list:
                value = [v.lower() for v in values]
            else:
                value = [values.lower()] if values else []
        elif filter_ == "upper":
            if values and type(values) == list:
                value = [v.upper() for v in values]
            else:
                value = [values.upper()] if values else []
        elif filter_ == "strip":
            if values and type(values) == list:
                value = [v.strip() for v in values]
            else:
                value = [values.strip()] if values else []
        elif filter_ == "capitalize":
            if values and type(values) == list:
                value = [v.capitalize() for v in values]
            else:
                value = [values.capitalize()] if values else []
        elif filter_ == "titlecase":
            if values and type(values) == list:
                value = [v.title() for v in values]
            else:
                value = [values.title()] if values else []
        elif filter_ == "braces":
            if values and type(values) == list:
                value = ["{" + v + "}" for v in values]
            else:
                value = ["{" + values + "}"] if values else []
        elif filter_ == "parens":
            if values and type(values) == list:
                value = ["(" + v + ")" for v in values]
            else:
                value = ["(" + values + ")"] if values else []
        elif filter_ == "brackets":
            if values and type(values) == list:
                value = ["[" + v + "]" for v in values]
            else:
                value = ["[" + values + "]"] if values else []
        elif filter_ == "shell_quote":
            if values and type(values) == list:
                value = [shlex.quote(v) for v in values]
            else:
                value = [shlex.quote(values)] if values else []
        elif filter_ == "split":
            # split on delimiter
            delim = args
            if delim:
                new_values = []
                for v in values:
                    new_values.extend(v.split(delim))
                value = new_values
            else:
                value = values
        elif filter_ == "chop":
            # chop off characters from the end
            chop = args
            try:
                chop = int(chop)
            except ValueError:
                raise SyntaxError(f"Invalid value for chop: {chop}")
            value = [v[:-chop] for v in values] if chop else values
        elif filter_ == "chomp":
            # chop off characters from the beginning
            chomp = args
            try:
                chomp = int(chomp)
            except ValueError:
                raise SyntaxError(f"Invalid value for chomp: {chomp}")
            value = [v[chomp:] for v in values] if chomp else values
        elif filter_ == "autosplit":
            # try to split keyword strings automatically
            temp_values = [v.replace(",", " ") for v in values]
            temp_values = [v.replace(";", " ") for v in temp_values]
            value = []
            for val in temp_values:
                value.extend(val.split())
        # elif filter_.startswith("function:"):
        # value = self.get_template_value_filter_function(filter_, values)
        else:
            raise SyntaxError(f"Unhandled filter: {filter_}")
        return value


def get_template_help() -> List[Union[str, List]]:
    """Return help for template system as list of markdown strings or lists of lists"""
    # TODO: would be better to use importlib.abc.ResourceReader but I can't find a single example of how to do this
    help_file = pathlib.Path(__file__).parent / "template.md"
    with open(help_file, "r") as fd:
        md = [fd.read()]

    # process help from plugins
    help_texts = PM.hook.get_template_help()
    for help_text in help_texts:
        # help_text is an iterable of str, dicts, or lists or lists
        for item in help_text:
            if type(item) == str:
                # format str as it might be a docstring
                md.append("\n" + dedent(item).strip())
            elif isinstance(item, (list, tuple)):
                md.append(item)
            else:
                raise ValueError(f"Unhandled help item: {item}")
    return md


# def lol_to_md_table(lol: List[List]) -> str:
#     """Convert a dict to a markdown table; assumes each list is the same length"""
#     if not lol:
#         return ""
#
#     markdowntable = ""
#     # Make a string of all the keys in the first dict with pipes before after and between each key
#     markdownheader = "| " + " | ".join(map(str, lol[0])) + " |"
#     # Make a header separator line with dashes instead of key names
#     markdownheaderseparator = "|-----" * len(lol[0]) + "|"
#     # Add the header row and separator to the table
#     markdowntable += markdownheader + "\n"
#     markdowntable += markdownheaderseparator + "\n"
#     # Loop through the list of lists outputting the rows
#     for row in lol[1:]:
#         markdownrow = "".join("| " + str(col) + " " for col in row)
#         markdowntable += markdownrow + "|" + "\n"
#     print(f"{markdowntable=}")
#     return markdowntable
#
# def get_template_value_function(
#     self,
#     subfield,
# ):
#     """Get template value from external function"""

#     if "::" not in subfield:
#         raise ValueError(
#             f"SyntaxError: could not parse function name from '{subfield}'"
#         )

#     filename, funcname = subfield.split("::")

#     filename_validated = expand_and_validate_filepath(filename)
#     if not filename_validated:
#         raise ValueError(f"'{filename}' does not appear to be a file")

#     template_func = load_function(filename_validated, funcname)
#     values = template_func(self.photo, options=self.options)

#     if not isinstance(values, (str, list)):
#         raise TypeError(
#             f"Invalid return type for function {funcname}: expected str or list"
#         )
#     if type(values) == str:
#         values = [values]

#     # sanitize directory names if needed
#     if self.filename:
#         values = [sanitize_pathpart(value) for value in values]
#     elif self.dirname:
#         # sanitize but don't replace any "/" as user function may want to create sub directories
#         values = [sanitize_dirname(value, replacement=None) for value in values]

#     return values

# def get_template_value_filter_function(self, filter_, values):
#     """Filter template value from external function"""

#     filter_ = filter_.replace("function:", "")

#     if "::" not in filter_:
#         raise ValueError(
#             f"SyntaxError: could not parse function name from '{filter_}'"
#         )

#     filename, funcname = filter_.split("::")

#     filename_validated = expand_and_validate_filepath(filename)
#     if not filename_validated:
#         raise ValueError(f"'{filename}' does not appear to be a file")

#     template_func = load_function(filename_validated, funcname)

#     if not isinstance(values, (list, tuple)):
#         values = [values]
#     values = template_func(values)

#     if not isinstance(values, list):
#         raise TypeError(
#             f"Invalid return type for function {funcname}: expected list"
#         )

#     return values



""" utility functions for validating/sanitizing path components """

import pathvalidate

# Max filename length on macOS
MAX_FILENAME_LEN = 255

# Max directory name length on macOS
MAX_DIRNAME_LEN = 255


def sanitize_filepath(filepath):
    """sanitize a filepath"""
    return pathvalidate.sanitize_filepath(filepath, platform="macos")


def is_valid_filepath(filepath):
    """returns True if a filepath is valid otherwise False"""
    return pathvalidate.is_valid_filepath(filepath, platform="macos")


def sanitize_filename(filename, replacement=":"):
    """replace any illegal characters in a filename and truncate filename if needed

    Args:
        filename: str, filename to sanitize
        replacement: str, value to replace any illegal characters with; default = ":"

    Returns:
        filename with any illegal characters replaced by replacement and truncated if necessary
    """

    if filename:
        filename = filename.replace("/", replacement)
        if len(filename) > MAX_FILENAME_LEN:
            parts = filename.split(".")
            drop = len(filename) - MAX_FILENAME_LEN
            if len(parts) > 1:
                # has an extension
                ext = parts.pop(-1)
                stem = ".".join(parts)
                if drop > len(stem):
                    ext = ext[:-drop]
                else:
                    stem = stem[:-drop]
                filename = f"{stem}.{ext}"
            else:
                filename = filename[:-drop]
    return filename


def sanitize_dirname(dirname, replacement=":"):
    """replace any illegal characters in a directory name and truncate directory name if needed

    Args:
        dirname: str, directory name to sanitize
        replacement: str, value to replace any illegal characters with; default = ":"; if None, no replacement occurs

    Returns:
        dirname with any illegal characters replaced by replacement and truncated if necessary
    """
    if dirname:
        dirname = sanitize_pathpart(dirname, replacement=replacement)
    return dirname


def sanitize_pathpart(pathpart, replacement=":"):
    """replace any illegal characters in a path part (either directory or filename without extension) and truncate name if needed

    Args:
        pathpart: str, path part to sanitize
        replacement: str, value to replace any illegal characters with; default = ":"; if None, no replacement occurs

    Returns:
        pathpart with any illegal characters replaced by replacement and truncated if necessary
    """
    if pathpart:
        pathpart = (
            pathpart.replace("/", replacement) if replacement is not None else pathpart
        )
        if len(pathpart) > MAX_DIRNAME_LEN:
            drop = len(pathpart) - MAX_DIRNAME_LEN
            pathpart = pathpart[:-drop]
    return pathpart



#!/usr/bin/env python

import os.path
import sys

from setuptools import find_packages, setup

PACKAGE_NAME = "autofile"

if sys.version_info < (3, 7, 0):
    sys.stderr.write(f"ERROR: You need Python 3.7 or later to use {PACKAGE_NAME}.\n")
    exit(1)

# we'll import stuff from the source tree, let's ensure is on the sys path
sys.path.insert(0, os.path.dirname(os.path.realpath(__file__)))

# read the contents of README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

about = {}
with open(
    os.path.join(this_directory, PACKAGE_NAME, "_version.py"),
    mode="r",
    encoding="utf-8",
) as f:
    exec(f.read(), about)

setup(
    name=PACKAGE_NAME,
    version=about["__version__"],
    description="Use templates to automatically move files into directories",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Rhet Turnbull",
    author_email="rturnbull+git@gmail.com",
    url="https://github.com/RhetTbull/autofile",
    project_urls={"GitHub": "https://github.com/RhetTbull/autofile"},
    download_url="https://github.com/RhetTbull/autofile",
    packages=find_packages(exclude=["tests", "utils"]),
    license="License :: OSI Approved :: MIT License",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: MacOS X",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: End Users/Desktop",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=[
        "click==8.0.1,<9.0.0",
        "cloup>=0.11.0,<0.12.0",
        "osxmetadata>=0.99.34,<1.0.0",
        "pathvalidate>=2.5.0,<3.0.0",
        "pdfminer.six>=20211012",
        "pluggy>=1.0.0,<=2.0.0",
        "pyobjc-core>=7.3,<8.0",
        "pyobjc-framework-Cocoa>=7.3,<8.0",
        "pyobjc-framework-CoreServices>=7.3,<8.0",
        "pyobjc-framework-FSEvents>=7.3,<8.0",
        "python-docx>=0.8.11,<1.0.0",
        "rich>=10.11.0,<11.0.0",
        "textX>=2.3.0,<3.0.0" "yaspin>=2.1.0",
        "tinytag>=1.6.0,<2.0.0",
        "yaspin>=2.1.0,<3.0.0",
    ],
    python_requires=">=3.7",
    entry_points={
        "console_scripts": [
            f"{PACKAGE_NAME}={PACKAGE_NAME}.cli:cli",
        ]
    },
    include_package_data=True,
)



""" plugin for autofile to return mdls data"""

import datetime
import plistlib
import subprocess
from typing import Dict, Iterable, List, Optional

import autofile
from autofile.datetime_utils import datetime_naive_to_utc

FIELDS = {
    "{mdls}": "Get metadata attributes for file as returned by mdls command; use in form '{mdls:ATTRIBUTE}', for example, '{mdls:kMDItemContentType}'",
}

CACHED_MDLS_DATA = {}


@autofile.hookimpl
def get_template_help() -> Iterable:
    text = """
    `{mdls:ATTRIBUTE}` returns the value of the metadata ATTRIBUTE as returned by the macOS `mdls` command. 
    For example, `{mdls:kMDItemContentType}` returns the content type of the file, e.g. `public.python-script` or `public.mp3` 
    and `{mdls:kMDItemKind}` returns a description of file type, e.g. `Python Script` or `MP3 Audio`.
    """
    fields = [["Field", "Description"], *[[k, v] for k, v in FIELDS.items()]]
    return ["**macOS Metadata Fields**", fields, text]


@autofile.hookimpl
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """lookup value for file dates

    Args:
        field: template field to find value for.

    Returns:
        The matching template value (which may be None).
    """
    if field != "mdls":
        return None

    if not subfield:
        raise ValueError(
            "{mdls} requires a subfield in form mdls:subfield, e.g. {mdls:kMDItemKind}"
        )

    global CACHED_MDLS_DATA
    if filepath in CACHED_MDLS_DATA:
        mdls = CACHED_MDLS_DATA[filepath]
        value = mdls.get(subfield)
        return value if type(value) == list else [value]

    mdls = load_mdls_data(filepath)
    CACHED_MDLS_DATA[filepath] = mdls
    value = mdls.get(subfield)
    return value if type(value) == list else [value]


def load_mdls_data(filepath: str) -> Dict:
    """load mdls data for file

    Args:
        filepath: path to file to load mdls data for

    Returns:
        dict of mdls data
    """
    cmd = ["mdls", "-plist", "-", filepath]
    proc = subprocess.run(cmd, capture_output=True)
    if proc.returncode != 0:
        raise subprocess.CalledProcessError(proc.returncode, cmd)
    mdls = plistlib.loads(proc.stdout)

    # plist returns naive datetime, should be UTC, not local timezone so convert if necessary
    for key, value in mdls.items():
        if isinstance(key, datetime.datetime):
            mdls[key] = datetime_naive_to_utc(value)
        else:
            mdls[key] = str(value)

    return mdls



"""exiftool template plugin for autofile"""

import datetime
from typing import Iterable, List, Optional

import autofile
from autofile.datetime_formatter import DateTimeFormatter
from autofile.exiftool import ExifToolCaching

FIELDS = {
    "{exiftool}": "Format: '{exiftool:GROUP:TAGNAME}'; use exiftool (https://exiftool.org) to extract metadata, "
    "in form GROUP:TAGNAME or TAGNAME, from image. "
    "E.g. '{exiftool:Make}' to get camera make, or {exiftool:IPTC:Keywords} to extract keywords. "
    "See https://exiftool.org/TagNames/ for list of valid tag names.  Group name is optional (e.g. EXIF, IPTC, etc) "
    "but if specified, should be the same as used in `exiftool -G`, e.g. '{exiftool:EXIF:Make}'. "
    "exiftool must be installed in the path to use this template field (https://exiftool.org/).",
}


DATETIME_ATTRIBUTES = {
    "date": "ISO date, e.g. 2020-03-22",
    "year": "4-digit year, e.g. 2021",
    "yy": "2-digit year, e.g. 21",
    "month": "Month name as locale's full name, e.g. December",
    "mon": "Month as locale's abbreviated name, e.g. Dec",
    "mm": "2-digit month, e.g. 12",
    "dd": "2-digit day of the month, e.g. 22",
    "dow": "Day of the week as locale's full name, e.g. Tuesday",
    "doy": "Julian day of year starting from 001",
    "hour": "2-digit hour, e.g. 10",
    "min": "2-digit minute, e.g. 15",
    "sec": "2-digit second, e.g. 30",
    "strftime": "Apply strftime template to date/time. Should be used in form "
    + "{created.strftime,TEMPLATE} where TEMPLATE is a valid strftime template, e.g. "
    + "{created.strftime,%Y-%U} would result in year-week number of year: '2020-23'. "
    + "If used with no template will return null value. "
    + "See https://strftime.org/ for help on strftime templates.",
}

DATETIME_SUBFIELDS = list(DATETIME_ATTRIBUTES.keys())


@autofile.hookimpl
def get_template_help() -> Iterable:
    text = """
    The `{exiftool}` template uses the third-party exiftool app (https://exiftool.org) to extract metadata from photo and video files.

    It must be used with one or more subfields which are exiftool tags, for example: `{exiftool:EXIF:Make}` for camera make, 
    or `{exiftool:IPTC:Keywords}` for keywords. The exiftool Group name (e.g. `IPTC`) is optional. 

    There are two derived subfields: `created` and `modified` which represent the created date or the modified date, respectively. 
    These subfields are datetime values and you can access the various attributes of the datetime by using an
    attribute name following a period, e.g. `{exiftool:created.year}` for the 4-digit year. 

    The following attributes are supported:

    """
    fields = [["Field", "Description"], *[[k, v] for k, v in FIELDS.items()]]
    attributes = [
        ["Attribute", "Description"],
        *[[k, v] for k, v in DATETIME_ATTRIBUTES.items()],
    ]
    return ["**Photo and Video Files**", fields, text, attributes]


@autofile.hookimpl
def get_template_value(
    filepath: str, field: str, subfield: str, default: List[str]
) -> Optional[List[Optional[str]]]:
    """lookup value for os.stat values for filepath

    Args:
        field: template field to find value for.

    Returns:
        The matching template value (which may be None).
    """
    if "{" + field + "}" not in FIELDS:
        return None

    exiftool = ExifToolCaching(filepath)
    exifdict = exiftool.asdict(normalized=True)
    exifdict_no_groups = exiftool.asdict(tag_groups=False, normalized=True)
    exifdict = exifdict.copy()
    exifdict.update(exifdict_no_groups)

    if not subfield:
        raise ValueError(f"subfield not specified for {field}")

    tag = subfield.lower()
    tag_subfield = None
    if "." in tag:
        tag, tag_subfield = tag.split(".")

    values = []
    if tag == "created":
        date = get_created_date(exiftool)
        if date:
            values = [str(date)]
    elif tag == "modified":
        date = get_modified_date(exiftool)
        if date:
            values = [str(date)]
    elif tag in exifdict:
        values = exifdict[tag]
        values = [values] if not isinstance(values, list) else values
        values = [str(v) for v in values]

        # "(Binary data " below is hack workaround for "(Binary data 0 bytes, use -b option to extract)" error that happens
        # when exporting video with keywords on Photos 5.0 / Catalina
        values = [v for v in values if not v.startswith("(Binary data ")]

    if values and tag_subfield:
        # handle datetime formatting
        if tag not in ["created", "modified"]:
            raise ValueError(
                "date/time formatting can only be used with created or modified subfields"
            )

        if tag_subfield not in DATETIME_SUBFIELDS:
            raise ValueError(f"Invalid value {tag_subfield} for date/time formatter")

        if tag_subfield == "strftime":
            if default:
                try:
                    values = [
                        datetime.datetime.fromisoformat(v).strftime(default[0])
                        for v in values
                    ]
                except:
                    raise ValueError(f"Invalid strftime template: '{default}'")
            else:
                value = None
        else:
            values = [
                getattr(
                    DateTimeFormatter(datetime.datetime.fromisoformat(v)), tag_subfield
                )
                for v in values
            ]

    return values


def get_created_date(exiftool):
    """Get created date from EXIF data or None"""

    data = exiftool.asdict()
    for tag in [
        "Composite:SubSecDateTimeOriginal",
        "Composite:DateTimeCreated",
        "QuickTime:CreationDate",
        "QuickTime:CreateDate",
        "EXIF:DateTimeOriginal",
        "EXIF:CreateDate",
        "IPTC:DateCreated",
        "XMP-xmp:CreateDate",
        "XMP-photoship:DateCreated",
    ]:
        if tag in data:
            return exiftool_date_to_datetime(data[tag])
    else:
        return None


def get_modified_date(exiftool):
    """Get modified date from EXIF data or None"""

    data = exiftool.asdict()
    for tag in [
        "Composite:SubSecModifyDate",
        "EXIF:ModifyDate",
        "QuickTime:ModifyDate",
        "XMP-prism:ModificationDate",
    ]:
        if tag in data:
            return exiftool_date_to_datetime(data[tag])
    else:
        return None


def exiftool_date_to_datetime(date: str) -> datetime.datetime:
    """Convert EXIF date to datetime.datetime"""

    """ exiftool can produce date/times in a variety of formats:
        with TZ offset: "2021:08:01 21:51:33-07:00",
        no TZ offset: "2019:07:27 17:33:28",
        no time: "2019:04:15",
        subsecond, no TZ offset: "2019:04:15 14:40:24.86",
        subsecond, TZ offset: "2019:04:15 14:40:24.86-04:00",
  """

    date_time_formats = [
        "%Y:%m:%d %H:%M:%S%z",
        "%Y:%m:%d %H:%M:%S",
        "%Y:%m:%d %H:%M:%S.%f",
        "%Y:%m:%d %H:%M:%S.%f%z",
        "%Y:%m:%d",
    ]

    for dt_format in date_time_formats:
        try:
            parsed_date = datetime.datetime.strptime(date, dt_format)
        except ValueError as e:
            pass
        else:
            return parsed_date
    else:
        raise ValueError(
            f"Could not parse date format: {date} does not match {date_time_formats}"
        )
